{ ====================================================================
The SFShellView class

Copyright 2014 Roelf Toxopeus

SwiftForth version.
SFShellView as contenView for a NSWindow.
Behaves (hopefully) as a Terminal.app console.
Last: 17 October 2020 at 23:33:48 CEST  -rt
==================================================================== }

{ --------------------------------------------------------------------
SFShellView = shellview

SFShellView -- NSTextView subclass, see the NSResponder class for what
to implement or override. A SFShellView class can be used like any
other NSView subclass as a contentView for a NSWindow instance.
SFShellView instance uses NSFileHandles for I/O. These can be provided
by pipes, using NSPipe instances. See shellview-console.f

The shellview instance will be notified if there's something to read:
/READNOTIFY -- install read notification for given shellview instance
-READNOTIFY -- remove it from given shellview

SFShellView methods:
SVOut SVIn -- instance variables, containing the I/O handles.
These are used internaly.

initWithIn:Out: -- initialiser, taking NSFileHandles for I/O

dealloc -- called when releasing the shellview instance.

keyDown: -- what to do when a key pressed in shellview, this is used
to send text to the output NSFileHandle.

mouseDown: -- action when mousebutton pressed. For now, ignores any
buttonpress.

handleNotification: -- action for received notifications, this is
used to read text from the input NSFileHandle and put it in the
shellview.

FILTER-IN -- used inside the notificationhandler. Used to massage
input. Can be extended to catch many more stuff of course.

---
Forth:
SHELLVIEW -- create a new shellview instance with the provided I/O
NSFileHandles.

-SHELLVIEW -- remove observer from Notificationcenter release the
NSFileHandles and finaly release the shellview.

WINDOWFORSHELL -- create and bind a new shellview to a scrollview
window.

Note: pressing a key will not echo the character directly in the
shellview. keyDown: intercepts the process. The character is send
without echoing to Forth and ACCEPT echoes the character back to the
shellview. Hitting return causes the Forth respons of course.

To test this, you can send a character string to Forth, where ACCEPT
echoes the chars. It's similar to JAMming text in to the input stream.

: ECHOES ( a n shellview -- )
   >R
   NSData @dataWithBytes:length:
   Z" SVOut" R> @IVAR @writeData: drop ;

s\" 2 3 + .\r" myshellview @ echoes  ( \r simulates hitting return )

This can be of help when implementing a MPW or Sibley editor like
console.
-------------------------------------------------------------------- }

/FORTH
DECIMAL

COCOA.FRAMEWORK

\ LACKING NSTextView COCOACLASS NSTextView
COCOACLASS NSData
COCOACLASS NSFileHandle
COCOACLASS NSNotificationCenter

COCOA: @writeData: ( NSDataRef -- ret )                                       \ NSFileHandle
COCOA: @dataWithBytes:length: ( string length -- NSDataRef )                  \ NSData
COCOA: @bytes ( -- addr )                                                     \ NSData
COCOA: @initWithData:encoding: ( nsdataref encoding -- nsstringref )          \ NSString
COCOA: @characters ( -- NSStringRef )                                         \ NSEvent
COCOA: @charactersIgnoringModifiers ( -- NSStringRef )                        \ NSEvent
COCOA: @defaultCenter ( -- sharednoticenter )                                 \ NSNotificationCenter
COCOA: @addObserver:selector:name:object: ( ref sel nstringref obj -- ret )   \ NSNotificationCenter
COCOA: @removeObserver:name:object: ( ref nsstringref obj -- ret )            \ NSNotificationCenter
COCOA: @object ( -- obj )                                                     \ NSNotification
COCOA: @userInfo ( -- ref )                                                   \ NSNotification
COCOA: @objectForKey: ( key -- obj )                                          \ NSArray
COCOA: @dealloc ( -- ret )                                                    \ NSObject

GLOBAL: NSFileHandleReadCompletionNotification                        \ nsstring constant, use @
GLOBAL: NSFileHandleNotificationDataItem                              \ nsstring constant, use @

1 CONSTANT NSASCIIStringEncoding

\ --------------------------------------------------------------------
\ utillities:

: ReadOnMain ( NSFileHandleRef -- )   >R Z" readInBackgroundAndNotify" @SELECTOR 0 YES R> FORMAIN DROP ;

: /READNOTIFY ( NSShellViewRef -- )
   Z" SVIn" OVER @IVAR >R                                 \ the object: SVIn NSFileHandleRef
                                                          \ observer, the shellview instance
   Z" handleNotification:" @SELECTOR                      \ selector
   NSFileHandleReadCompletionNotification @               \ name
   R@                                                     \ object
   NSNotificationCenter @defaultCenter                    \ the receiver
   @addObserver:selector:name:object: DROP                \ add to notificationcenter
   R> ReadOnMain                                          \ start it
;

: -READNOTIFY ( NSShellViewRef -- )
   Z" SVIn" OVER @IVAR >R                                 \ the object: SVIn NSFileHandleRef
                                                          \ observer, the shellview instance
   NSFileHandleReadCompletionNotification @               \ name
   R>                                                     \ object
   NSNotificationCenter @defaultCenter                    \ the receiver   
   @removeObserver:name:object: DROP ;

\ --------------------------------------------------------------------
\ NSTextView subclass, see NSResponder for what to implement/override.

\ the method implementations:

\ this version gets the NSFileHandles for reading and writing directly, no conversion needed
FORTHCALL: *initWithIn:Out: ( rec sel fhin fhout -- shellViewRef )
   _PARAM_0 @init >R                        \ call super's init method
   _PARAM_2 Z" SVIn" R@ !IVAR               \ and store them in the instance variables
   _PARAM_3 Z" SVOut" R@ !IVAR
   R@ /READNOTIFY
   R> ;                                     \ return new shellView instance

: INITTYPES ( -- zaddr )   Z" @@:ii" ;

FORTHCALL: *keyDown: ( rec sel NSEventRef -- ret )
   _PARAM_2 @characters @UTF8String            \ cstring, alternative: @charactersIgnoringModifiers
   ZCOUNT NSData @dataWithBytes:length:        \ wrap in NSData
   Z" SVOut" _PARAM_0 @IVAR @writeData: 0 ;    \ receiver is shellview instance. write to its filehandle out
   
: GETEVENTTYPES ( -- a )   0" v@:@" ;

FORTHCALL: *mouseDown: ( rec sel NSEventRef -- ret ) ( we ignore it, we don't move the textcursor! ) 0 ;

: IOTYPES ( -- zaddr )   Z" @" ;

\ the class:
NSTextView NEW.CLASS SFShellView

IOTYPES Z" SVOut" SFShellView ADD.IVAR

IOTYPES Z" SVIn"  SFShellView ADD.IVAR

*initWithIn:Out: INITTYPES     Z" initWithIn:Out:" SFShellView ADD.METHOD

*keyDown:        GETEVENTTYPES Z" keyDown:"        SFShellView ADD.METHOD

*mouseDown:      GETEVENTTYPES Z" mouseDown:"      SFShellView ADD.METHOD

(* we're not done yet, but will add last method later *)

SFShellView ADD.CLASS

\ the shelview in window:
COCOA: @initWithIn:Out: ( fhin fhout -- shellViewRef )

: SHELLVIEW ( fhin fhout -- SFShellView:ref )
   SFShellView @alloc @initWithIn:Out: DUP >R
   NO  R@ @setRichText: DROP
   YES R@ @setVerticallyResizable: DROP
   NO  R@ @setHorizontallyResizable: DROP
   NSViewWidthSizable R@ @setAutoresizingMask: DROP
   YES R> @textContainer @setWidthTracksTextView: DROP
;

: -SHELLVIEW ( shellview -- )
   DUP -READNOTIFY
   Z" SVIn"  OVER @IVAR @release DROP
   Z" SVOut" OVER @IVAR @release DROP
   @release DROP ;

: (WINDOWFORSHELL) ( fhin fhout wptr4 -- shellview )
   +W.REF @ >R
   SHELLVIEW SCROLLVIEW 2DUP                   \ new shell- and scrollview
   @setDocumentView: DROP                      \ connect shellview to scrollview
   R@ @setContentView: DROP                    \ set scrollview to window's contentview
   0 R@ @makeKeyAndOrderFront: DROP            \ make window the event receiver
   DUP R> @makeFirstResponder: DROP            \ make shellview the text receiver
   POCKET !
;

: WINDOWFORSHELL ( fhin fhout wptr4 -- shellview )
   3 0 ['] (WINDOWFORSHELL) PASS POCKET @ ;

\ --------------------------------------------------------------------
\ Add notification handler method to the SFShellView class
\ using SFShellView instance as the observer...
\ Yes, we can add methods to an existing class.
\ An example of using the passed along receiver parameter ( _PARAM_0 ) in an ObjC method.

: >SHELLVIEW ( NSStringRef SFShellViewRef -- )   >R Z" insertText:" @SELECTOR SWAP YES R> FORMAIN DROP ;

\ $082008   CONSTANT SF.BS   \ in sf simple accept sends bs space bs on backspace key event
\ $08082008 CONSTANT SF.ESC  \ in sf simple accept sends bs sequences to erase the whole line

CREATE "CSI" 27 C, CHAR [ C, 0 C,

: FILTER-IN ( SFShellViewRef NSDataRef -- )
   SWAP >R
   DUP @bytes
   DUP C@ 8 = IF ( one or more bs sequences )
      DROP ( bytes )
      @length 3 = IF Z" deleteBackward:" ELSE  Z" deleteToBeginningOfLine:" THEN
      @SELECTOR 0 YES R> FORMAIN DROP
      EXIT
   THEN

   OVER @length "CSI" 2 SEARCH IF ( looking for escape sequence )
      DROP ( length result )
      2 CHARS + C@ [CHAR] H = IF ( clear screen ? )
         Z" selectAll:" @SELECTOR 0 YES R@ FORMAIN DROP
         Z" deleteToMark:" @SELECTOR 0 YES R> FORMAIN DROP
         DROP ( nsdataref )
         EXIT
      THEN
   ELSE  2DROP ( search result )
   THEN
   
   ( neither BS or CLS )
   NSASCIIStringEncoding                              \ ascii data
   NSString @alloc @initWithData:encoding:            \ nstringref
   DUP R> >SHELLVIEW @release DROP ;
      
: READ>SHELLVIEW ( shellviewref notificationref -- )
   ( *) TUCK   
   NSFileHandleNotificationDataItem @                 \ key, what to look for
   SWAP @userInfo                                     \ infoblurb
   @objectForKey:                                     \ data
   FILTER-IN
   ( *) @object ReadOnMain                            \ do read again, @object returns the object causing the notification: an NSFileHandleRef
;

FORTHCALL: *handleNotification: ( rec sel notification -- ret )   _PARAM_0 _PARAM_2 READ>SHELLVIEW 0 ;

: NOTIFICATIONTYPES ( -- a )   z" v@:@" ;

*handleNotification: NOTIFICATIONTYPES Z" handleNotification:" SFShellView ADD.METHOD

CR .( shellview loaded)

\\ ( eof )