{ ====================================================================
Tread priority setting

Copyright 2003 Roelf Toxopeus

SwiftForth version
Originates CarbonMacForth
In pursuit of controlling the Mach scheduler, especially in time critical
situations, like multimedia processing.
Last: 27 October 2019 at 16:11:50 CET     -rt
==================================================================== }

{ --------------------------------------------------------------------
Idea is, that a thread can tell the scheduler to hold off until further
notice. This 'until further notice' isn't infinite, there _are_ limits.
Tricky part is the timing stuff.

In Terminal:  $ sysctl kern.clockrate

Info in:
Darwin ( Kernel, mach scheduling ) developer documentation
url: file:///Developer/Documentation/DarwinCoreOS/Conceptual/KernelProgramming/scheduler/
https://developer.apple.com/library/archive/documentation/Darwin/Conceptual/KernelProgramming/scheduler/scheduler.html#//apple_ref/doc/uid/TP30000905-CH211-BEHJDFCA
Header files found in /usr/include directory
mach/mach_init.h mach/thread_policy.h
mach/mach_time.h sched.h etc. etc.

More info:
https://flylib.com/books/en/3.126.1.80/1/   scheduling chapter 7.4
points to
https://opensource.apple.com/source/xnu/xnu-1456.1.26/osfmk/kern/sched_prim.c.auto.html
std_quantum (10,000 μs), the standard timeslicing quantum

min_std_quantum (250 μs), the smallest remaining quantum

min_rt_quantum (50 μs), the smallest real-time computation

max_rt_quantum (50 ms), the largest real-time computation

sched_tick_interval (1000 >> SCHED_TICK_SHIFT ms)

The policy info block for the realtime policy flavor has 4 32b fields.
From Apple docs
field 0 = period: nominal amount of time between separate processing
          arrivals, 0 indicates that there is no inherent periodicity.
field 1 = computation: nominal amount of computation time needed
          during a separate processing arrival. Don't make this too
          short, 50 microsecs minimum rt computation!
field 2 = constraint: This is the maximum amount of real time that
          may elapse from the start of a separate processing arrival
          to the end of computation for logically correct functioning.
          Must be (>= computation).
          Note that latency = (constraint - computation).
field 3 = preemptible: This indicates that the computation may be
          interrupted, subject to the constraint specified above.

Calculate the time periods in Absolute Time units.
see mac/utils/abstime2.f
On this machine Absolute Time units equal nonoseconds.

?REALTIME -- Check realtime computation for min max realtime interval.
SOLO -- tell kernel scheduler this thread goes for realtime with set
timetable. This could mean hold off preempting.
-SOLO -- tell kernel scheduler this thread goes normal...

Note:
Doesn't look to be very useful according the timer tests,
try it with time critical interrupt stuff.
See mac/sound/coreaudio/noise/synchronator/xsine.f
for a realworld example. It certainly works!

Realtime alternative in system/spawn4.f
/REALTIME ['] whatever SPAWN  /SPAWN
-------------------------------------------------------------------- }

/FORTH
DECIMAL

FRAMEWORK System.framework

FUNCTION: mach_thread_self ( -- mach_port_t )
FUNCTION: thread_policy_set ( thread flavor policy_info count -- return )

\ using an extern info block for now
CREATE TIMETABLE 4 CELLS /ALLOT

50000 CONSTANT RT-MIN     \ abstime tics = nanosecs -> 50 musecs
50000000 CONSTANT RT-MAX  \ abstime tics = nanosecs -> 50 msecs

: ?REALTIME ( n -- n )
   DUP RT-MIN < ABORT" realtime slice too short !"
   DUP RT-MAX > ABORT" realtime slice too large !" ;
   
\ Like SINGLE in some multitasking Forth systems.
: SOLO ( -- ret )
   mach_thread_self              \ thread
   2                             \ 2 = THREAD_TIME_CONSTRAINT_POLICY
   TIMETABLE                     \ policy info address
   4                             \ #32b in info block
   thread_policy_set ;

: -SOLO ( -- return )
   mach_thread_self              \ thread
   1                             \ 1 = THREAD_STANDARD_POLICY
   0 >R RP@                      \ info address, needed?
   0                             \ no parameters
   thread_policy_set
   R> DROP ;

CR .( SOLO and -SOLO priority, check or DROP returns)

\\ ( eof)

\ --------------------test and usage--------------------

mac include utils/abstime2.f

1500000 4 * value undisturbed
variable accu
: tally     undisturbed 0 do 1 accu +! loop ;

atime tally atime 2swap d- drop ?realtime value atics/tally

atime 1000 ms atime 2swap d- drop constant atics/sec

\ fill in the time info block
timetable
0           over ! cell+      \ undefined tallies per sec
atics/tally over ! cell+      \ time it takes for our routine
atics/tally over ! cell+      \ no latency
0 swap !                      \ no interrups

: sally ( -- )   solo abort" no solo!" tally -solo drop ;

\\

\ you'll notice the difference between the examples when other applications
\ are running alongside MF, like iTunes, playing a cd, and Safari:

cr atime tally atime 2swap d- d. many
cr atime sally atime 2swap d- d. many

\ play a cd with iTunes, on my system the execution of ss will cause
: tt 10 0 do tally 100 ms loop ;
cr atime tt atime 2swap d- d. many
: ss solo drop tt -solo drop ;
cr atime ss atime 2swap d- d. many

\\

\ Another example in C
/* ================================================== */
/*
  Give chronyd real time priority so that time critical calculations
  are not pre-empted by the kernel.
*/

static int
set_realtime(void)
{
  /* https://developer.apple.com/library/ios/technotes/tn2169/_index.html */

  mach_timebase_info_data_t timebase_info;
  double clock2abs;
  thread_time_constraint_policy_data_t policy;
  int kr;

  mach_timebase_info(&timebase_info);
  clock2abs = ((double)timebase_info.denom / (double)timebase_info.numer) * NANOS_PER_MSEC;

  policy.period = 0;
  policy.computation = (uint32_t)(5 * clock2abs); /* 5 ms of work */
  policy.constraint = (uint32_t)(10 * clock2abs);
  policy.preemptible = 0;

  kr = thread_policy_set(
          pthread_mach_thread_np(pthread_self()),
          THREAD_TIME_CONSTRAINT_POLICY,
          (thread_policy_t)&policy,
          THREAD_TIME_CONSTRAINT_POLICY_COUNT);

  if (kr != KERN_SUCCESS) {
    LOG(LOGS_WARN, "Cannot set real-time priority: %d", kr);
    return -1;
  }
  return 0;
}

/* ================================================== */

void
SYS_MacOSX_SetScheduler(int SchedPriority)
{
  if (SchedPriority) {
    set_realtime();
  }
}
