(* ====================================================================
OS fixes

Copyright 2014 Roelf Toxopeus

SwiftForth version.
Some work-arounds for (temprary?) OS issues.
These are for 10.11 El Capitan and up.
Last: 15 October 2020 at 11:14:07 CEST  -rt
==================================================================== *)

(( --------------------------------------------------------------------
Since OSX 10.11 El Capitan, many GUI calls better be executed on the main
thread to avoid some issues. This includes some NIB handling.
Looks like things changed a bit again in macOS 10.12 Sierra. Settings should
be done on the main thread as well to avoid flickering while resizing a window.
Experimenting will clarify.

So far the following has been observed:
Main thread prefered -> initialisation, running modal, property adjustments,
setting title etc.
Also all contentview (NSView and subs) stuff should be PASSed or FORMAINed.

Secundary thread allowed -> opening, hiding, showing, closing, moving,
sizing, queries (@frame) 

The official Thread Programming Guide from Apple regarding the windows:
https://developer.apple.com/library/content/documentation/Cocoa/Conceptual/Multithreading/ThreadSafetySummary/ThreadSafetySummary.html#//apple_ref/doc/uid/10000057i-CH12-123351-BBCFIIEB

It does not talk about the changes in OSX 10.11 El Capitan and later. AFAIK we
have always done it like they tell us to do. So something fishy...
Notice that SHOW.NIB works from secundary thread in many cases. But when you use
auto layout in your nibs, it should be run from the main thread.  Only the
showWindow: message needs to run from the main thread.

Probably the best practice is to identify the part which should run
on the main thread and only run that on the main thread. Most of the
time this is just one message. But sometimes it's a few, mixed with
others and they're passed in one go like the fix for CREATE.WINDOW.

For changing window background colours @setBackgroundColor: should run
on the main thread. But in modal event loops it should _not_ do so.
Any PASS in a modal event loop will hang. So this method is not re-
defined to run on the main loop, allowing it to be used in modal run
loops. But the specific window background colour setters will be PASSed.

To get an idea what's happening, try the following (open Console.app as well)
and observe:
nswindow @alloc @init
dup @orderFront: drop

Initialize Window on Main thread using PASS and switch to the main thread via
PAUSE and a MS instruction to have the initializer executed.
Upon return to the Forth thread show it.
Failing to do that, window resizing will flicker for all windows!
It appears you have to init _all_ things on the main thread to avoid resize
flickering later on in _all_ existing windows ...

/NIB              -- redefined using mainthread
SHOW.NIB          -- redefined using mainthread

SET.WTITLE        -- redefined using mainthread
SET.WTRANSPARENCY -- redefined using mainthread
SET.WBACKGROUND   -- redefined using mainthread

DEFAULT.BG BBE.BG WHITE.BG RED.BG BLACK.BG -- redefined using mainthread

(CREATE.WINDOW)   -- only creates window on mainthread
CREATE.WINDOW     -- creates and initialises window on main thread 
ADD.WINDOW        -- redefined using new CREATE.WINDOW
-------------------------------------------------------------------- ))

/FORTH
DECIMAL

\ --elcapitan--
\ Main thread redefinitions for NSNIB
: (/NIB) ( nib -- )   /NIB POCKET ! ;
: /NIB ( nib -- flag )   1 0 ['] (/NIB) PASS POCKET @ ;

: SHOW.NIB ( za -- windowcontroller )
   >NIBPATH
   DUP >4THSTRING FILE-STATUS NIP IF @release DROP 0 EXIT THEN
   NIBWINDOW.CONTROLLER 0 OVER  2 0 ['] @showWindow: PASS ;

\ --sierra--
\ Main thread redefinitions for setting window properties
: SET.WTITLE ( 0string wptr4 -- )    2 0 ['] SET.WTITLE PASS ;

: SET.WTRANSPARENCY ( wptr4 -- ) ( F: alpha -- )    1 1 ['] SET.WTRANSPARENCY PASS ;

: SET.WBACKGROUND ( wptr4 -- ) ( F: red green blue alpha -- )    1 4 ['] SET.WBACKGROUND PASS ;

\ --mojave--
: DEFAULT.BG ( wptr4 -- )   1 0 ['] DEFAULT.BG PASS ;

: BBE.BG ( wptr4 -- )   1 0 ['] BBE.BG PASS ;

: WHITE.BG ( wptr4 -- )   1 0 ['] WHITE.BG PASS ;

: RED.BG ( wptr4 -- )   1 0 ['] RED.BG PASS ;

: BLACK.BG ( wptr4 -- )   1 0 ['] BLACK.BG PASS ;

\ --elcapitan--
\ Main thread redefinitions for NSWindow using the redefined words.
: (CREATE.WINDOW) ( wptr4 -- )
   DUP WINDOW.ARGS   
   NSWindow @alloc @initWithContentRect:styleMask:backing:defer:
\ default opaque setting is YES, so need to chage it to change transparency!
   NO OVER @setOpaque: DROP
   SWAP +W.REF ! ;

: CREATE.WINDOW ( wptr4 -- )
	DUP 1 0 ['] (CREATE.WINDOW) PASS 100 MS PAUSE     \ --elcapitan--
	DUP +W.TITLE @ COUNT DROP OVER SET.WTITLE         \ now on main thread
   DUP +W.TRANSPARENCY SF@ DUP SET.WTRANSPARENCY     \ now on main thread 32b !!!
   DUP +W.BACKGROUND CG4@ SET.WBACKGROUND ;          \ now on main thread
 
: ADD.WINDOW ( wptr4 -- )
   DUP +W.REF @ IF DUP CLOSE.WINDOW THEN
   DUP CREATE.WINDOW
   SHOW.WINDOW ;

cr .( OS fixes loaded)

\\ (eof )
