{ ====================================================================
Exception frame decode

Copyright 2011  FORTH, Inc.

A diagnostic dump of the kernel's exception frame is provided here.
The system's .CATCH is extended to display a decode of this info when
a system exception causes a THROW.

======================================================================
Overview of SwiftForth exception handling

KERNEL

The kernel implements CATCH and THROW, common to all platforms
(Windows, Linux, macOS):

SwiftForth\src\kernel\except.f (this file)

Each platform adds a handler to "trap" the system exceptions:

SwiftForth/src/kernel/linux/trap.f
SwiftForth/src/kernel/macos/trap.f
SwiftForth/src/kernel/win32/trap.f

Each of them registers its exception handler, named <EXCEPTION> with
the system. In Linux and macOS, this is done with multiple calls to
sigaction(), one for each "signal" were willing to handle. In
Windows, theres a single call to AddVectoredExceptionHandler() and
its up to the handler to determine if it will process the exception
or pass it along to be handled by another member of chain the of
registered handlers.

If the exception is for us, we receive from the OS the user context
for the exception. We capture the CPU registers from that context and
save it in the TRACEBACK buffer. We also save some return stack and
Forth data stack (as pointed to by ESP and EBP in the CPU register
context) for debug display.

We replace the EIP within the context with the address of THROWBACK
and return, indicating with our return status in EAX that we have
handled the exception and that the OS should resume execution from the
stored user context. The OS will then return back to the user-level
code using the stored context, which now will continue execution at
THROWBACK. And that just does IOR_EXCEPT THROW.

THROW will also capture return and data stacks in the same way if the
throw code passed to it is not IOR_EXCEPT. So no matter how you got a
THROW (system exception or user code calling THROW directly), you can
dump out the stacks for debug in a consistent way.

ELECTIVES

Up in the IDE world, we extend the systems .CATCH display to perform
.THROW, which will dump the captured CPU context and stacks if the
throw code is IOR_EXCEPT. All three platforms do this the same way in
this shared source file:

SwiftForth\src\ide\exception.f

The system-specific exception decode and mapping of the CPU context
registers is supplied in that platforms exception-map.f.

In Windows, we add the display of both kinds of debug info (THROW,
which shows stacks but not registers) and system exception (which
shows all of it) in a dialog box:

SwiftForth\src\ide\win32\exception-dialog.f

==================================================================== }

PACKAGE ERROR-HANDLERS

?( Signal exception traceback dump)

{ --------------------------------------------------------------------
Traceback display

The kernel's exception handler captures error frame info into the
TRACEBACK buffer.  This consists of three things:

1) The CPU registers from the user context passed to the exception handler
2) The return stack pointed by ESP in the user context
3) The Forth data stack pointed to by EBP in the user context

The display words below decode the error frame data. The register
contents, the top 20 items on the data stack, and the top 20 entities
on the return stack are presented.

The return stack is used to instantiate many different temporary
things in SwiftForth; each of these is decoded separately and
displayed in some pretty fashion. This means that some items on the
return stack aren't displayed in full, but a much better picture of
the overall error reference is made.

Each token to be decoded requires a known return stack address to
identify it, so each entity has a code-space label.

Local entities don't have persistent names, but local objects have a
discernible class and so the class names are presented.  Data pools
defined by R-ALLOC and R-BUF aren't displayed, but their existence and
size are.  Local variables are likewise not displayed but their
existence is reported.
-------------------------------------------------------------------- }

: DICTIONARY? ( addr -- flag )
   ORIGIN 1+   UP0 @REL [ H UP@ - ] LITERAL + @  WITHIN ;

: (ID.') ( addr -- )
   DUP (.')  DUP ORIGIN <  IF  DROP  C" <unknown>"  THEN
   ( a nfa)  COUNT 2DUP TYPE SPACE  + 1+ -
   ?DUP IF  DUP 0> IF ." +" THEN  . THEN ;

: ID.' ( addr -- )
   DUP DICTIONARY? IF  (ID.') EXIT THEN
   +ORIGIN  DUP DICTIONARY? IF  (ID.') EXIT  THEN  DROP ;

: @REG ( r -- x )
   CELLS TRACEBACK + @ ;

: @RS ( i -- x )
   #TRACE_REGS + CELLS  TRACEBACK + @ ;

: .LOCAL-OBJECT ( i1 -- i2 )
   DUP DUP 1+ @RS CELL / + >R 3 + @RS  ID.'  R> ;

: .LOCAL-OBJECTS ( i1 -- i2 )
   BEGIN
      DUP #TRACE_RSTACK U< WHILE
      DUP @RS IOR_EXCEPT <> WHILE
      .LOCAL-OBJECT
   REPEAT THEN ;

: .RSTACK-SKIP ( i1 -- i2 n )
   DUP @RS  REG_ESP CELLS TRACEBACK + @ - CELL /
   TUCK SWAP - 1- ;

: .REGX ( n -- )
   CASE
      0 OF  ." EAX "  REG_EAX  ENDOF
      1 OF  ." EBX "  REG_EBX  ENDOF
      2 OF  ." ECX "  REG_ECX  ENDOF
      3 OF  ." EDX "  REG_EDX  ENDOF
      4 OF  ." ESI "  REG_ESI  ENDOF
      5 OF  ." EDI "  REG_EDI  ENDOF
      6 OF  ." EBP "  REG_EBP  ENDOF
      7 OF  ." ESP "  REG_ESP  ENDOF
      8 OF  ." EFL "  REG_EFL  ENDOF
      9 OF  ." EIP "  REG_EIP  ENDOF
   DROP  14 SPACES  EXIT  ENDCASE
   @REG H.8  SPACE SPACE ;

: .DS ( offset -- )
   #TRACE_REGS + #TRACE_RSTACK + CELLS
   TRACEBACK + @ H.8  SPACE SPACE ;

: .RS ( i1 -- i2 )
   #TRACE_RSTACK OVER < ?EXIT
   DUP 1+ SWAP @RS
   DUP H.8 SPACE  -ORIGIN CASE
      'R-ALLOC OF ." <<r-alloc>> "  .RSTACK-SKIP .  ENDOF
      'R-BUF   OF ." <<r-buf>> "    .RSTACK-SKIP .  ENDOF
      'LSPACE  OF ." <<locals>> "   .RSTACK-SKIP .  1+  ENDOF
      'OSPACE  OF ." <<object>> "   .LOCAL-OBJECTS  2+  ENDOF
   DUP ID.'  ENDCASE ;

PUBLIC

: .TRACEBACK1 ( -- )     \ Dump TRACEBACK info including registers
   ."  at "  REG_EIP @REG DUP H.8  SPACE  ID.'
   CR ." Registers     Dstack    esp+ Rstack  "
   0  20 0 DO
      CR  I .REGX  I .DS  DUP CELLS 4 H.0 SPACE .RS
   LOOP DROP ;

: .TRACEBACK2 ( -- )    \ Dump TRACEBACK stacks only
   CR ." Dstack    esp+ Rstack  "
   0  20 0 DO
     CR  I .DS  DUP CELLS 4 H.0 SPACE .RS
   LOOP DROP ;

{ ----------------------------------------------------------------------
Display exception frames

IOR_EXCEPT is added to the (THROW) switch to properly decode the
"signal" exception events.

.THROW extends the system .CATCH behavior to include .TRACEBACK debug
info for IOR_EXCEPT events.
---------------------------------------------------------------------- }

: EXCEPTION-DECODE ( -- addr u )
   EXCEPTION# @ EXCEPTION-TYPE ;

[+SWITCH (THROW)
   IOR_EXCEPT RUN: EXCEPTION-DECODE ;
SWITCH]

: .THROW ( ior -- )
   DUP ERRORMSG  IOR_EXCEPT = IF  .TRACEBACK1  THEN ;

' .THROW IS .CATCH

END-PACKAGE

\\ =====================================================================

{ ----------------------------------------------------------------------
Tests for traceback dump
---------------------------------------------------------------------- }

: ZZ ( -- )   7 >R  33 R-ALLOC  9 >R  0 @  HERE . ;
: ZZZ ( -- )   3 >R  4 >R  ZZ  HERE . ;

: YY ( -- )   7 >R  R-BUF  9 >R  0 @ HERE . ;
: YYY ( -- )   3 >R  4 >R  YY HERE . ;

[UNDEFINED] POINT [IF]

CLASS POINT
   VARIABLE x
   VARIABLE y
END-CLASS

[THEN]

: XX ( -- )
   7 >R  [OBJECTS  POINT MAKES PT  OBJECTS]
   $11111111 PT X !  $55555555 PT Y !
   9 >R  0 @  HERE . ;

: WWW ( -- )   7 >R
   1 2 3 LOCALS| A B C | 9 >R  0 @  HERE . ;
