{ ====================================================================
String functions

Copyright 2001  FORTH, Inc.  All rights reserved

Windows and Linux require many strings to contain embedded control
characters. Typically <tab> and <newline>, but others may be required.

     \a   BEL (alert, ASCII 7)
     \b   BS (backspace, ASCII 8)
     \e   ESC (not in C99, ASCII 27)
     \f   FF (form feed, ASCII 12)
     \l   LF (ASCII 10)
     \m   CR/LF pair (ASCII 13, 10)
     \n   implementation-dependent newline, e.g. CR/LF, LF, or LF/CR.
     \q   double-quote (ASCII 34)
     \r   CR (ASCII 13)
     \t   HT (tab, ASCII 9)
     \v   VT (ASCII 11)
     \z   NUL (ASCII 0)
     \"   double-quote (ASCII 34)
     \x<hexdigit><hexdigit>
          The resulting character is the conversion of these two
          hexadecimal digits. An ambiguous conditions exists if \x
          is not followed by two hexadecimal characters.
     \\   backslash itself
     \    An ambiguous condition exists if a \ is placed before any
          character, other than those defined above.
          SwiftForth aborts with a warning message.

==================================================================== }

?( String functions)

{ ------------------------------------------------------------------------
Embedded escape sequences

GETCH takes buffer address and length, gets a one-byte character from
it, and returns the remaining buffer, length, and character.

PUTCH adds a character to the string being built at HERE.

\BAD complains about the ambiguous condition noted above.
------------------------------------------------------------------------ }

?( ... Embedded escape sequences)

PACKAGE STRING-TOOLS

THROW#
   S" Unexpected quoted character"   >THROW ENUM IOR_QUOTED
TO THROW#

: GETCH ( addr1 u1 -- addr2 u2 char )
   1 /STRING  OVER 1- C@ ;

: PUTCH ( char -- )
   HERE COUNT + C!  1 HERE C+! ;

: \BAD ( addr1 u1 char -- )   IOR_QUOTED THROW ;

: HEXDIG ( char -- n )
   UPPER DUP [CHAR] 9 > 7 AND - [CHAR] 0 - ;

: \HEX ( a n -- a+2 n-2 )
   GETCH HEXDIG 4 LSHIFT >R  GETCH HEXDIG  R> OR  PUTCH ;

[SWITCH ADDCH \BAD ( addr1 u1 char -- addr2 u2 )
   CHAR a RUN:  [CTRL] G PUTCH ;                \ alert (bell)
   CHAR b RUN:  [CTRL] H PUTCH ;                \ backspace
   CHAR e RUN:  27 PUTCH ;                      \ escape
   CHAR f RUN:  [CTRL] L PUTCH ;                \ form feed
   CHAR l RUN:  [CTRL] J PUTCH ;                \ line feed
   CHAR m RUN:  13 PUTCH  10 PUTCH ;            \ cr,lf (platform-independent)
   CHAR n RUN:  <EOL> COUNT HERE APPEND ;       \ eol (platform-dependent)
   CHAR q RUN:  [CHAR] " PUTCH ;                \ alternate double-quote
   CHAR " RUN:  [CHAR] " PUTCH ;                \ double-quote
   CHAR r RUN:  [CTRL] M PUTCH ;                \ cr
   CHAR t RUN:  [CTRL] I PUTCH ;                \ htab
   CHAR v RUN:  [CTRL] K PUTCH ;                \ vtab
   CHAR x RUN:  \HEX ;                          \ hex digits
   CHAR z RUN:  0 PUTCH ;                       \ nul
   CHAR 0 RUN:  0 PUTCH ;                       \ alternative nul
   CHAR \ RUN:  [CHAR] \ PUTCH ;                \ backslash
SWITCH]

PUBLIC

: FORMAT ( addr1 len -- addr2 )
   0 HERE C!
   BEGIN ( a n)
      DUP 0> WHILE  GETCH
         DUP [CHAR] \ = IF
            DROP GETCH ADDCH
         ELSE  PUTCH  THEN
   REPEAT 2DROP  HERE ;

PRIVATE

: "\PARSE ( -- addr )
   0 HERE C!  /SOURCE DUP >R
   BEGIN  DUP 0> WHILE
      GETCH CASE
         [CHAR] " OF  NEGATE >IN +!  0  ENDOF
         [CHAR] \ OF  GETCH ADDCH  ENDOF
      DUP PUTCH  ENDCASE
   REPEAT  R> SWAP - >IN +!  DROP  HERE ;

: "\QPAD ( -- )
   $80 QHEAD +!  "\PARSE COUNT QPAD PLACE ;

PUBLIC

: UPLACE ( caddr n addr -- )
   OVER IF
      -ROT  256 R-ALLOC >R  R@ PLACE
      R> COUNT BOUNDS DO
         I C@ OVER W!  2+
      LOOP 0 SWAP W!
   ELSE
      3DROP
   THEN ;

{ --------------------------------------------------------------------
Compile strings

S" ( -- a n )
Z" ( -- zstr )
U" ( -- ustr )
C" ( -- a )

Parse a string till the next quote. If interpreting, the string
is placed in a temporary buffer and the appropriate string reference
is returned on the stack. If compiling, the string is laid down
in memory and when executed will produce the correct string reference.

S\"
Z\"
U\"
C\"

Same as above, but allow escape characters.

,"
,Z"
,U"

Parse a string, allocate memory for it, and place it in the dictionary.

,\"
,Z\"
,U\"

Same as above, but allow escape characters.
-------------------------------------------------------------------- }

?( ... String compilers)

: ,U" ( -- )   [CHAR] " WORD  ( c-addr) COUNT U, ;
: ,Z" ( -- )   [CHAR] " WORD  ( c-addr) COUNT Z, ;

: ,\" ( -- )   "\PARSE  C@ 1+ ALLOT 0 C, ALIGN ;
: ,Z\" ( -- )   "\PARSE ( c-addr) COUNT Z, ;
: ,U\" ( -- )   "\PARSE ( c-addr) COUNT U, ;

: S\" ( -- )
   STATE @ IF  POSTPONE (S")  ,\"  EXIT THEN  "\QPAD
   QPAD COUNT ;  IMMEDIATE

: C\" ( -- )
   STATE @ IF  POSTPONE (C")  ,\"  EXIT THEN  "\QPAD
   QPAD ;  IMMEDIATE

: Z\" ( -- )
   STATE @ IF  POSTPONE (Z")  ,\"  EXIT THEN  "\QPAD
   0 QPAD COUNT + C!  QPAD 1+ ;  IMMEDIATE

: .\" ( -- )
   POSTPONE S\"  POSTPONE TYPE ;  IMMEDIATE

END-PACKAGE

{ --------------------------------------------------------------------
String substitution

REPLACE changes old string to new string if found in source string.
-------------------------------------------------------------------- }

?( ... String substitution)

PACKAGE STRING-TOOLS

CREATE STRBUF   256 ALLOT

PUBLIC

: SUBST ( source len old len new len -- result len )
   LOCALS| nl n pl p |   0 STRBUF C!
   BEGIN   DUP 0> WHILE
      2DUP p pl SEARCH WHILE ( s l m l)
         ROT OVER - >R ROT R> STRBUF APPEND
         n nl STRBUF APPEND
         pl /STRING
   REPEAT 2SWAP STRBUF APPEND THEN
   2DROP STRBUF COUNT ;

: REPLACE ( new len source len old len -- result len )
   DUP>R  2OVER 2>R
   SEARCH IF    ( new l a n)
      DUP 2R> ROT - STRBUF PLACE
      R> /STRING  2SWAP STRBUF APPEND  STRBUF APPEND
   ELSE         ( new l a n)
      2R> STRBUF PLACE  R>DROP 2DROP 2DROP
   THEN  STRBUF COUNT ;

: REPLACE-CHAR ( source n oldch newch -- )
   2SWAP BEGIN ( old new a n)
      DUP WHILE
      FOURTH SCAN
      DUP IF 3DUP DROP C! THEN
   REPEAT 2DROP 2DROP ;

: SPLITSTR ( source n pattern n -- remainder n beginning n )
   2>R OVER SWAP BEGIN
      DUP 0> WHILE
      OVER 2R@ TUCK  COMPARE(CS) WHILE  \ Compare using current case sensitivity
      1 /STRING
   REPEAT THEN
   ROT THIRD OVER - 2R> 2DROP ;

PRIVATE

: -BLANKS ( addr n -- addr n )
   BL SKIP -TRAILING ;

: <BLANKS> ( addr n addr -- )   >R
   S"  " R@ PLACE  -BLANKS  R@ APPEND  S"  " R> APPEND ;

PUBLIC

: MARKED ( addr n pat n -- )
   256 R-ALLOC 256 R-ALLOC LOCALS| src pat |
   pat <BLANKS>  src <BLANKS>
   src COUNT BEGIN ( a n)
      DUP 0> WHILE
      pat COUNT SPLITSTR TYPE
      DUP 0> WHILE
         SPACE BRIGHT
         pat COUNT -BLANKS TYPE
         NORMAL
         pat C@ 1- /STRING
   REPEAT THEN 2DROP ;

: ZNEXT ( z -- z )   ZCOUNT + 1+ ;

PRIVATE

{ ------------------------------------------------------------------------
Standard string substitution

UNESCAPE, REPLACES, and SUBSTITUTE are part of the Forth-2012 standard.

UNESCAPE replaces each `%' character in the input string c-addr1 u1 by
two `%' characters. The output is represented by c-addr2 u2. The
buffer at c-addr2 shall be big enough to hold the unescaped string. An
ambiguous condition occurs if the resulting string will not fit into
the destination buffer (c-addr2).

SUBST-WORDLIST is a wordlist used to hold substitution names and
replacement text.

REPLACES sets the string c-addr1 u1 as the text to substitute for the
substitution named by c-addr2 u2. If the substitution does not exist
it is created. The program may reuse the buffer c-addr1 u1 without
affecting the definition of the substitution.

+OBUF appends a character to the output buffer
>OBUF appends a string to the output buffer

GET-NAME takes a source string c-addr1 len1 pointing at a leading
delimiter and returns the remainder of the source string c-addr2 len2
after the trailing delimiter and the name string c-addr3 len3. If the
string does not contain a trailing delimiter, len3 is returned zero.

REPLACE-NAME processes the substitution candidate c-addr len. Returns
true if found, 0 if not found.

SUBSTITUTE performs substitution on the string c-addr1 u1 placing the
result at string c-addr2 u3, where u3 is the length of the resulting
string. An error occurs if the resulting string will not fit into
c-addr2 u2 or if c-addr2 is the same as c-addr1. The return value n is
positive or 0 on success and indicates the number of substitutions
made. A negative value for n indicates that an error occurred, leaving
c-addr2 u3 undefined. Negative values of n are implementation defined
except for values in table 9.1 THROW code assignments.

------------------------------------------------------------------------ }

WORDLIST CONSTANT SUBST-WORDLIST

: +SUBST ( c-addr len -- c-addr )
   SUBST-WORDLIST (WID-CREATE)  LAST @ >CREATE !
   HERE 256 ALLOT 0 OVER C! ;

: ?SUBST ( c-addr len -- xt flag | 0 )
   SUBST-WORDLIST SEARCH-WORDLIST ;

PUBLIC

: REPLACES ( c-addr1 u1 c-addr2 u2 -- )
   2DUP ?SUBST IF  NIP NIP EXECUTE      \ get existing buffer address
   ELSE  +SUBST  THEN  PLACE ;          \ copy as counted string

PRIVATE

2VARIABLE OBUF          \ Destination buffer current length and address
VARIABLE #SUBST         \ Number of substitutions made

: +OBUF ( char -- )
   OBUF 2@ + C!  1 OBUF C+! ;

: >OBUF ( c-addr len -- )
   BOUNDS ?DO  I C@ +OBUF  LOOP ;

: GET-NAME ( c-addr1 len1 -- c-addr2 len2 c-addr3 len3 )
   1 /STRING  2DUP '%' SCAN DUP IF
      >R DROP  2DUP R> - TUCK
   2>R 1+ /STRING  2R>  THEN ;

: REPLACE-NAME ( c-addr len -- flag )
   2DUP ?SUBST IF
      NIP NIP EXECUTE COUNT >OBUF  TRUE
      ELSE  '%' +OBUF  >OBUF  '%' +OBUF  FALSE
   THEN ;

PUBLIC

: SUBSTITUTE ( c-addr1 u1 c-addr2 u2 -- c-addr2 u3 n )
   R-BUF  R@ 0 OBUF 2!  0 #SUBST !  2SWAP
   BEGIN  DUP 0>  WHILE
      OVER C@ '%' <> IF  OVER C@ +OBUF 1 /STRING
         ELSE  OVER 1+ C@  '%' = IF  '%' +OBUF  2 /STRING
            ELSE  GET-NAME DUP IF
               REPLACE-NAME IF  1 #SUBST +!  THEN
              ELSE  2DROP  '%' +OBUF
   THEN THEN THEN  REPEAT 2DROP
   OBUF @ < IF  R> DROP 0 -1
   ELSE  R> OVER OBUF @ CMOVE  OBUF @ #SUBST @  THEN ;

: UNESCAPE ( c-addr1 u1 c-addr2 -- c-addr2 u2 )
   DUP 2SWAP OVER + SWAP ?DO
      I C@ [CHAR] % = IF  [CHAR] % OVER C! 1+  THEN
   I C@  OVER C! 1+  LOOP  OVER - ;

PRIVATE

: (.SUBST) ( nfa -- flag )
   CR  DUP COUNT TYPE ."  --> "  NAME> EXECUTE COUNT TYPE  -1 ;

PUBLIC

: .REPLACES ( -- )      \ Dump the replacement list
   ['] (.SUBST) SUBST-WORDLIST TRAVERSE-WORDLIST ;

END-PACKAGE
