{ ====================================================================
Command line editor

Copyright 2008  FORTH, Inc.  All rights reserved

This file implements enhanced console version of ACCEPT for command-
line input, editing, and history.
==================================================================== }

?( Command line editor)

DEFER TOGGLE-BROWSER

PACKAGE ACCEPTOR

{ --------------------------------------------------------------------
ACCEPT  The built-in function is very simple -- it responds to only
the backspace, escape, and carriage return controls.

Backspace does the obvious, moving the cursor backwards until it
reaches the beginning of the line.

Escape kills all characters in the current line.

Carriage return ends the input, leaving the number of characters
accepted on the stack.
-------------------------------------------------------------------- }

{ --------------------------------------------------------------------
Line Buffering

Text is managed in a buffer LBUF of packed strings.  Each new
string is added to the beginning by PUSHLINE, and the oldest strings
will simply fall off the end.
-------------------------------------------------------------------- }

8192 CONSTANT |LBUF|

0 VALUE LBUF

: /LBUFFER ( -- )    |LBUF| 256 + ALLOCATE THROW TO LBUF
   LBUF |LBUF| 256 + ERASE ;

/LBUFFER

{ --------------------------------------------------------------------
Lines in the line buffer

PUSHLINE places string addr u at the head of the buffer if it isn't
already there.

-LBUF is true if the string starting at the address is not complete in
LBUF. This is important because strings can be pushed off the end of
the buffer.

NTH-LINE returns addr and len of the Nth line in the buffer.

#LBUF is how many complete strings are in the buffer.
-------------------------------------------------------------------- }

: PUSHLINE ( addr u -- )
   DUP 0> IF
      2DUP LBUF COUNT COMPARE(CS) IF
         LBUF                           \ a n l
         OVER LBUF + 1+                 \ a n l l'
         |LBUF| FOURTH 1+ -             \ a n l l' n'
         CMOVE>                         \ a n
         LBUF PLACE  EXIT
   THEN THEN  2DROP ;

: -LBUF ( a -- flag )
   COUNT +  LBUF |LBUF| + < NOT ;

: NTH-LINE ( n -- addr n )
   LBUF SWAP   0 ?DO ( a)
      DUP -LBUF IF 0 UNLOOP EXIT THEN
      COUNT +
   LOOP COUNT ;

: #LBUF ( -- n )
   0  LBUF BEGIN ( a)
      DUP C@ WHILE
      DUP -LBUF 0= WHILE
      COUNT +  SWAP 1+ SWAP
   REPEAT THEN DROP ;

{ --------------------------------------------------------------------
THELINE holds the current consideration line when reviewing the
   contents of the line buffer.

CLAMP-LINE translates line numbers into guaranteed valid range.

GET-LINE sets the consideration line up and returns it.

+LINE moves +/- n lines from the current consideration line.

.HISTORY prints the top N items in the line buffer.

TYPE-KEYBOARD-HISTORY prints the entire line buffer.
-------------------------------------------------------------------- }

VARIABLE THELINE        \ 0..n where n is variable
VARIABLE RECALLED?
VARIABLE RECALLING?

: CLAMP-LINE ( n -- n )
   #LBUF ?DUP IF
      >R BEGIN ( n)
         DUP 0 R@ WITHIN NOT WHILE
         DUP 0< IF R@ + ELSE R@ - THEN
      REPEAT R> DROP
   THEN ;

: GET-LINE ( n -- addr n )
   CLAMP-LINE  DUP THELINE !  NTH-LINE ;

: +LINE ( n -- addr n )
   THELINE @ + GET-LINE ;

PUBLIC

: .HISTORY ( n -- )
   #LBUF MIN
   1- 0 MAX 0 SWAP ?DO
      CR I NTH-LINE TYPE
   -1 +LOOP CR ;

: TYPE-KEYBOARD-HISTORY ( -- )   #LBUF .HISTORY ;

PRIVATE

{ --------------------------------------------------------------------
Initialize, save
-------------------------------------------------------------------- }

: SAVE-LBUF ( -- )
   #LBUF IF  S" HOME" FIND-ENV IF
      R-BUF  2DUP R@ PLACE  S" /.SwiftForth_history" R@ APPEND
      R> COUNT R/W CREATE-FILE 0= IF
         LBUF #LBUF 0 DO  COUNT 2DUP 4 PICK WRITE-LINE DROP +  LOOP DROP
   DUP CLOSE-FILE DROP  THEN DROP  THEN 2DROP  THEN ;

: RESTORE-LBUF ( -- )
   S" HOME" FIND-ENV IF  R-BUF  2DUP R@ PLACE  S" /.SwiftForth_history" R@ APPEND
      R> COUNT R/O OPEN-FILE 0= IF  LBUF  BEGIN
         DUP 1+ 256 3 PICK READ-LINE DROP WHILE
      OVER C!  COUNT +  REPEAT DROP  0 SWAP C!  DUP CLOSE-FILE DROP
   THEN DROP  THEN 2DROP ;

:ONENVLOAD ( -- )
   /LBUFFER  RESTORE-LBUF ;

:ONENVEXIT ( -- )
   SAVE-LBUF ;

{ --------------------------------------------------------------------
Input pattern matching is convenient. Look thru history for a string
whose beginning matches the given string.
-------------------------------------------------------------------- }

CREATE PATTERN   256 /ALLOT

: /COMPLETE ( a n # -- a n # )
   THIRD PATTERN COUNT TUCK COMPARE(CS)
   PATTERN C@ 0=  OR  IF  THIRD OVER PATTERN PLACE
   -1 THELINE !  THEN ;

: REPATTERN ( a n # -- a n # )
   THIRD OVER PATTERN PLACE
   -1 THELINE ! ;

{ --------------------------------------------------------------------
Line Editing

The functions associated with editing a line involve moving a cursor
around, redisplaying text, and overstriking or inserting new characters
into the buffer.

The functions with the A- prefix implement the accept editor.  They
(mostly) expect the address and length (A N) of the buffer to edit,
and the current position (#) of the cursor in the buffer.

Input is finished when the cursor position is negative.
-------------------------------------------------------------------- }

CREATE INSERTING  -1 ,     \ 0=overstrike, nonzero=insert

: A-TOGGLE-INSERT ( a n # -- a n # )
   INSERTING @ 0= INSERTING ! ;

{ --------------------------------------------------------------------
TAIL returns the original string plus the addr/len of its tail.

-------------------------------------------------------------------- }

: TAIL ( a n # -- a n # a' n' )
   3DUP /STRING ;

{ --------------------------------------------------------------------
A-SPREAD inserts a single character gap in the given string at the
   offset # .  The end of the string goes to the bit bucket.

A-TUCK removes the character at offset # .  The end of string is
   filled with a blank.
-------------------------------------------------------------------- }

: A-SPREAD ( a n # -- )
   TUCK  - 1-  0 MAX  -ROT  + DUP 1+  ROT CMOVE> ;

: A-TUCK ( a n # -- )   3DUP
   TUCK  - 1-  0 MAX  -ROT  + DUP 1+ SWAP ROT CMOVE
   DROP + 1-  BL SWAP C! ;

: A-LEFT ( a n # -- a n # )
   DUP IF 1- 8 EMIT THEN ;

: A-LEFTDEL ( a n # -- a n # )
   DUP IF  1-  8 EMIT  BL EMIT  8 EMIT  THEN ;

: A-RIGHT ( a n # -- a n # )
   2DUP > IF  THIRD OVER + C@ EMIT 1+ THEN ;

: A-LEFTWORD ( a n # -- a n # )
   A-LEFT BEGIN
      DUP WHILE
      THIRD OVER + C@ BL = WHILE
      A-LEFT
   REPEAT THEN
   BEGIN
      DUP WHILE
      THIRD OVER + C@ BL <> WHILE
      A-LEFT
   REPEAT THEN
   DUP IF A-RIGHT THEN  REPATTERN ;

: A-RIGHTWORD ( a n # -- a n # )
   THIRD THIRD -TRAILING NIP >R ( max)  BEGIN
      DUP R@ < WHILE
      THIRD OVER + C@ BL <> WHILE
      A-RIGHT
   REPEAT THEN
   BEGIN
      DUP R@ < WHILE
      THIRD OVER + C@ BL = WHILE
      A-RIGHT
   REPEAT THEN
   R> DROP REPATTERN ;

: A-HOME ( a n # -- a n 0 )
   BEGIN DUP WHILE A-LEFT REPEAT ;

: A-END ( a n # -- a n # )
   THIRD THIRD -TRAILING NIP BEGIN ( a n # x)
      2DUP <> WHILE
      2DUP < IF    >R A-RIGHT R>
             ELSE  >R A-LEFT  R> THEN
   REPEAT DROP ;

: A-CR ( a n # -- a n # )
   3DUP A-END -ROT 2DROP MAX  NIP -1 ;

: A-TYPE< ( a n -- )
  ?DUP IF
     GET-XY  2SWAP TYPE  0 OVER AT-XY  AT-XY  EXIT
   THEN DROP ;

: A-DELETE ( a n # -- a n # )
   3DUP A-TUCK  TAIL A-TYPE< ;

: A-BACKSPACE ( a n # -- a n # )
   DUP IF
      INSERTING @ IF
         A-LEFT A-DELETE
      ELSE
         A-LEFTDEL  TAIL DROP BL SWAP C!
      THEN
   THEN  REPATTERN ;

: A-CANCEL ( a n # -- a n # )
   BEGIN  DUP WHILE  A-LEFT  REPEAT DROP
   2DUP BLANK  GET-XY 2>R  2DUP TYPE  2R> AT-XY  0 ;

: A-CLEAR ( a n # -- a n # )
   DUP >R  PAGE  A-HOME  A-END  A-HOME
   R> 0 ?DO A-RIGHT LOOP ;

: A-OVERSTRIKE ( a n # char -- a n # )
   >R  2DUP > IF ( a n #)
      THIRD OVER + R@ SWAP C!  R@ EMIT  1+
   THEN R> DROP ;

: A-INSERT ( a n # char -- a n # )
   >R  2DUP > IF
      3DUP A-SPREAD  TAIL DROP R@ SWAP C!  R@ EMIT  1+
      TAIL -TRAILING A-TYPE<
   THEN R> DROP ;

: A-REPLACE ( a n # a n -- a n # )
   2>R A-CANCEL 2R> BOUNDS ?DO
      I C@ A-OVERSTRIKE
   LOOP  A-HOME A-END ;

: A-TRANSPOSE ( a n # -- a n # )
   THIRD THIRD -TRAILING NIP  ?DUP -EXIT
   2DUP  1+ 1 WITHIN IF  DROP EXIT  THEN
   OVER 1 = OVER 1 = AND IF DROP EXIT  THEN
   OVER =  IF  A-LEFT  THEN  THIRD OVER +
   1- DUP 2C@ SWAP  2DUP 2>R  ROT 2C!
   A-LEFT  2R> EMIT EMIT  2+ ;

: A-RECALL-UP ( a n # -- a n # )
   #LBUF -EXIT   1 +LINE A-REPLACE ;

: A-RECALL-DOWN ( a n # -- a n # )
   #LBUF -EXIT  -1 +LINE A-REPLACE ;

: A-MATCH-LINE?  ( a n # -- a n # flag )  \ True = Match
   THELINE @ NTH-LINE   PATTERN COUNT  ROT MIN TUCK COMPARE(CS) 0= ( Match? ) ;

: A-LINE-REPLACE  ( a n # -- a n # )
   THELINE @ NTH-LINE  A-REPLACE ;

: A-COMPLETE  ( a n # -- a n # )
   /COMPLETE
   #LBUF 0
   ?DO  THELINE @ 1+ #LBUF OVER U> AND THELINE !
      A-MATCH-LINE? IF  A-LINE-REPLACE  UNLOOP EXIT  THEN
   LOOP ;

: -A-COMPLETE ( a n # -- a n # )
   /COMPLETE  THELINE @ 0 MAX THELINE !
   #LBUF 0
   ?DO  #LBUF 1- THELINE @ ?DUP IF  1- NIP THEN THELINE !
      A-MATCH-LINE?
      IF  A-LINE-REPLACE  UNLOOP EXIT  THEN
   LOOP ;

: A-UP-ARROW   ( a n # -- a n # )
   DUP RECALLING? @ NOT AND IF
   A-COMPLETE  RECALLED? OFF  EXIT  THEN
   A-RECALL-UP  RECALLED? ON ;

: A-DN-ARROW   ( a n # -- a n # )
   DUP RECALLING? @ NOT AND IF
   -A-COMPLETE  RECALLED? OFF  EXIT  THEN
   A-RECALL-DOWN  RECALLED? ON ;

: A-SHOW-HIST  ( a n # -- a n # )  A-CANCEL  10 .HISTORY ;

: A-DEL-EOL  ( a n # -- a n # )
   R-BUF THIRD OVER R@ PLACE R> COUNT A-REPLACE ;

: A-APPEND-LAST  ( a n # -- a n # )
   A-DEL-EOL
   0 NTH-LINE THIRD /STRING 0 MAX BOUNDS
   ?DO  I C@ A-OVERSTRIKE  LOOP  REPATTERN ;

: A-LEFT-REPAT       ( a n # -- a n # )  A-LEFT      REPATTERN ;
: A-RIGHT-REPAT      ( a n # -- a n # )  A-RIGHT     REPATTERN ;
: A-END-REPAT        ( a n # -- a n # )  A-END       REPATTERN ;

: CHARACTER ( a n # char -- a n # )
   INSERTING @ IF A-INSERT ELSE A-OVERSTRIKE THEN ;

{ --------------------------------------------------------------------
F8/F9 shortcuts
-------------------------------------------------------------------- }

: (STUFF) ( a n # a n -- a n # )
   2DUP TYPE  >R >R  2DROP  R> OVER R@ CMOVE  R> -1 ;

: "BUILD" ( -- ) ( a n # -- a n # )
   OVER IF  A-CANCEL  THEN  S" INCLUDE build" (STUFF) ;

: "DEBUG" ( -- ) ( a n # -- a n # )
   OVER IF  A-CANCEL  THEN  S" INCLUDE debug" (STUFF) ;

{ --------------------------------------------------------------------
Line Editing Functions

Control characters mapped are:
   cr           end input
   bs           destroy last character typed
   esc          destroy current line
   tab          completion
   null         destroy current line, enter interpreter mode

Extended characters mapped are:
   del          delete character under the cursor
   left arrow   move cursor left
   right arrow  move cursor right
   home         move cursor to start of line
   end          move cursor to end of line
   ins          toggle insert mode
   up arrow     replace current line with most recent line
   down arrow   replace current line with oldest line
   ctrl home    display 10 most recent lines
-------------------------------------------------------------------- }

: A-QUIT ( a n # -- a n # )   DUP IF  7 EMIT EXIT  THEN  BYE ;

: BLANK? ( a n -- flag )      BL SKIP NIP 0= ;

: A-EOF ( a n # -- a n # )
  >R  2DUP BLANK?  R> SWAP IF A-QUIT ELSE A-DELETE THEN ;

: A-DEL-LEFTWORD ( a n # -- a n # )
  DUP >R  A-LEFTWORD  R> OVER - 0 ?DO A-DELETE LOOP ;

: E-STROKE ( a n # echar -- a n # )
   CASE
      VERASE OF A-BACKSPACE ENDOF
      [CTRL] H OF A-BACKSPACE ENDOF
      VKILL OF A-CANCEL ENDOF
      $1B OF A-CANCEL ENDOF             \ esc
      K-HOME OF A-HOME ENDOF
      [CTRL] A OF A-HOME ENDOF
      K-END OF A-END-REPAT ENDOF
      [CTRL] E OF A-END-REPAT ENDOF
      [CTRL] I OF A-COMPLETE ENDOF      \ tab
      [CTRL] M OF A-CR ENDOF            \ cr
      K-DELETE OF A-DELETE ENDOF
      K-LEFT OF A-LEFT-REPAT ENDOF
      K-RIGHT OF A-RIGHT-REPAT ENDOF
      K-UP OF A-UP-ARROW ENDOF
      K-DOWN OF A-DN-ARROW ENDOF
      [CTRL] P OF A-UP-ARROW ENDOF      \ bash up/down
      [CTRL] N OF A-DN-ARROW ENDOF
      K-CTRL_LEFT OF A-LEFTWORD ENDOF
      K-CTRL_RIGHT OF A-RIGHTWORD ENDOF
      K-INSERT OF A-TOGGLE-INSERT ENDOF
      K-F3 OF A-APPEND-LAST ENDOF       \ F3 - Append Last to EOL
      VEOF OF A-EOF ENDOF
      VWERASE OF A-DEL-LEFTWORD ENDOF
      [CTRL] B OF A-LEFT ENDOF
      [CTRL] F OF A-RIGHT ENDOF
      [CTRL] K OF A-DEL-EOL ENDOF
      [CTRL] L OF A-CLEAR  ENDOF        \ bash clear
      [CTRL] T OF A-TRANSPOSE ENDOF
      K-F8 OF  "BUILD"  ENDOF
      K-F9 OF  "DEBUG"  ENDOF
      DUP 32 256 WITHIN IF  CHARACTER REPATTERN DUP  THEN
   ENDCASE ;

: E-ACCEPT  ( a n -- n )
   -1 THELINE !  PATTERN OFF
   GET-SIZE DROP 1- MIN
   2DUP BLANK  0 BEGIN ( a n # )
      AKEY
      RECALLED? @ RECALLING? !  RECALLED? OFF
      E-STROKE
      RECALLED? @ RECALLING? !
   DUP 0< UNTIL DROP
   TUCK  PUSHLINE ;

' E-ACCEPT 'ACCEPT !	\ Reassign the console terminal personality's ACCEPT vector

END-PACKAGE
