{ ====================================================================
Generate portable executable turnkey program

Copyright 2001  FORTH, Inc.

PROGRAM <objfile> <icon>

  1) icon file name ends in .ico
  2) icon format must be single image no greater than 64k in size

======================================================================== }

?( Turnkey program generator)

PACKAGE PROGRAM-GENERATOR

{ ------------------------------------------------------------------------
Buffers
------------------------------------------------------------------------ }

MAX_PATH BUFFER: NAME_OBJ       \ Object filename
MAX_PATH BUFFER: NAME_ICO       \ Icon filename

{ ---------------------------------------------------------------------
Loader file header

These objects, pointers, and offsets are used to access the sf-loader
header image.

(FIND-HEADER) scans the loader header in memory for the given string.
Aborts if not found.

VALIDATE-HEADER checks that the image header starts with MH_MAGIC.
--------------------------------------------------------------------- }

PUBLIC

IMAGE_NT_HEADERS BUILDS PE-HEADER       \ PE header
IMAGE_SECTION_HEADER BUILDS .rsrc       \ PE resource section
IMAGE_SECTION_HEADER BUILDS .forth      \ Forth section header (named .data)
IMAGE_RESOURCE_DATA_ENTRY BUILDS icon1  \ App icon image pointer, size
IMAGE_RESOURCE_DATA_ENTRY BUILDS icon2  \ App icon group pointer, size
GRPICONDIR BUILDS icongroup             \ App icon group metadata (pointed to by icon2)

VARIABLE 'SF                            \ Address of SwiftForth image

PRIVATE

VARIABLE 'ICO                           \ Address of icon image

: (FIND-HEADER) ( addr1 len -- addr2 )
   'FILEIMAGE @ #LOADER @ 2SWAP SEARCH 0= ABORT" Loader section not found" DROP ;

PUBLIC

HEX

: SCAN-HEADER ( -- )
   S\" PE\0\0" (FIND-HEADER) PE-HEADER SET
   S\" \x89PNG" (FIND-HEADER) DUP 'ICO !
   10 - DUP icon2 SET  10 - icon1 SET
   S" .rsrc" (FIND-HEADER) .rsrc SET
   S" .data" (FIND-HEADER) .forth SET
   .forth PointerToRawData L@  #LOADER @
   <> ABORT" Loader size mismatch"
;

PRIVATE

{ ------------------------------------------------------------------------
Checksum image

CheckSumMappedFile calculates the checksum of the exectuable image
before we write it to the object file.

PIMAGE_NT_HEADERS IMAGEAPI CheckSumMappedFile(
  PVOID  BaseAddress,
  DWORD  FileLength,
  PDWORD HeaderSum,
  PDWORD CheckSum
)

CHECKSUM-IMAGE calls CheckSumMappedFile and returns the new checksum.
------------------------------------------------------------------------ }

LIBRARY Imagehlp.dll
FUNCTION: CheckSumMappedFile ( addr1 len addr2 addr3 -- x )

VARIABLE SUM0   \ Original checksum from CheckSumMappedFile
VARIABLE SUM1   \ Calculated checksum


PUBLIC

: CHECKSUM-IMAGE ( -- x )
   'FILEIMAGE @  #FILEIMAGE @           \ addr len for entire image about to be written to object file
   SUM0 SUM1 CheckSumMappedFile         \ Calculate the checksum, returned in SUM1
   0= ABORT" Can't checksum image"      \ Complain if checksum fails
   SUM1 @ ;                             \ Return new checksum

PRIVATE

{ --------------------------------------------------------------------
Icon in PE header

If the next string in the input stream ends in ".ico" we open it and
use it as the turnkey image's icon.

The icon file must be conform to the 6-byte header in ICON-FMT.

MAX-ICONSIZE returns the maximum size of the ICO image (less header).
The entire resource section for the sf-loader image contains a single
icon and its group (of 1, with ID 101 for the program).
-------------------------------------------------------------------- }

16 BUFFER: ICON-META

: /GRPICONDIR ( n -- )
   7 + -8 AND  'ICO @ +                 \ Address for new GRPICONDIR meta
   ICON-META OVER icongroup SET         \ Inialize icon meta data
   'FILEIMAGE @ -                       \ Offset to rawdata
   .rsrc PointerToRawData L@ -          \ Offset within section image
   .rsrc VirtualAddress L@ +            \ Convert to virtual address
   icon2 OffsetToData L! ;              \ Update group icon pointer

CREATE ICON-FMT
   0 W,  1 W,  1 W,     \ reserved=0; type=icon; exactly 1 image in file

: MAX-ICONSIZE ( -- n )
   .rsrc VirtualSize L@ 100 - ;

: READ-ICON ( -- )
   NAME_ICO C@ IF
      NAME_ICO COUNT R/O OPEN-FILE THROW >R
      R@ FILE-SIZE THROW DROP
      MAX-ICONSIZE > ABORT" ICO file too big"
      ICON-META 16 R@ READ-FILE THROW DROP      \ Read ICO file header meta data
      ICON-META 6 ICON-FMT OVER COMPARE         \ Validate first 3 words of ICO header
      ABORT" Unsupported icon format"
      'ICO @ MAX-ICONSIZE 2DUP ERASE
      R@ READ-FILE  R> CLOSE-FILE DROP  THROW
      DUP icon1 Size L!  /GRPICONDIR
   THEN ;

{ --------------------------------------------------------------------
Build file image

VALIDATE-HEADER does a quick sanity check to make sure fields are
where we expect them in the header.

BUILD-IMAGE reads in the PE header from sf-loader.bin, makes sure it's
in the expected format, updates the header size fields, appends the
code image, and updates the PE header checksum.
-------------------------------------------------------------------- }

[UNDEFINED] EXPORT: [IF]

0 CONSTANT #EXPORTS
: ,XDIR 2DROP ;
: DLL-HEADER ;

[ELSE]

: DLL-HEADER ( -- )
   DLL-ENTRY -ORIGIN #LOADER @ +IMG !           \ Set entry point offset at ORIGIN
   >XDIR 2@  PE-HEADER DataDirectorySize0 L!    \ Set PE header export table size
   PE-HEADER DataDirectoryVirtualAddress0 L!    \ Set PE header export table virtual address
   10000000 PE-HEADER ImageBase !               \ Default ImageBase for DLLs
   230E PE-HEADER Characteristics W!
   0 PE-HEADER DllCharacteristics W! ;          \ File and DLL characteristics

[THEN]

VARIABLE ?SEALED        \ True for "sealed" turnkey (RFU)

: BUILD-IMAGE ( -- )
   NAME_OBJ COUNT -PATH ,XDIR                   \ Build export table
   HERE -ORIGIN #XREF + 200 CHUNK DUP >R        \ Object image size
   S" %SwiftForth\bin\sf-loader.bin"
   READ-IMAGE  SCAN-HEADER                      \ Read the PE header into memory, scan header
   R@ PE-HEADER SizeOfInitializedData L+!       \ Add kernel size to SizeOfInitializedData
   R> .forth SizeOfRawData L!                   \ set SizeOfRawData in .forth section
   2 IS-CONSOLE? -  PE-HEADER Subsystem W!      \ Set Subsystem for program type (2=GUI, 3=console)
   READ-ICON                                    \ Overlay icon if .ico specified
   HERE DP0 !REL  HERE SAVE-XREF                \ Set DP0 for turnkey, set xref restore addr
   ORIGIN  #LOADER @ +IMG                       \ Source from origin, dest after PE header in object file image
   HERE -ORIGIN 2DUP 2>R CMOVE                  \ Copy SwiftForth into file image
   2R> + SWAP CMOVE                             \ Copy xref table to image
   #EXPORTS IF  DLL-HEADER  THEN                \ Set DLL header fields if there are exports
   CHECKSUM-IMAGE PE-HEADER Checksum L! ;       \ Set the new checksum in the header

DECIMAL

{ --------------------------------------------------------------------
Files
-------------------------------------------------------------------- }

: (PARSE-OBJ) ( -- )
   BL WORD COUNT +ROOT NAME_OBJ PLACE
   NAME_OBJ COUNT -PATH  [CHAR] . SCAN
   0= IF  #EXPORTS IF  S" .dll"  ELSE  S" .exe" THEN
   NAME_OBJ APPEND  THEN DROP ;

: (PARSE-ICO) ( -- )   >IN @
   BL WORD COUNT  2DUP 4 STRING/ S" .ico"
   COMPARE(NC) IF  2DROP >IN !  0 NAME_ICO C!
   ELSE  +ROOT NAME_ICO PLACE  DROP  THEN ;

{ --------------------------------------------------------------------
Program turnkey

PROGRAM is followed by the object file name and an option icon name.
It sets up the object and icon file names, builds the object file
image in extended memory, and writes it to the file.
-------------------------------------------------------------------- }

PUBLIC

: PROGRAM ( -- )
   (PARSE-OBJ)  (PARSE-ICO)  BUILD-IMAGE        \ Build the turnkey file memory image
   NAME_OBJ COUNT R/W CREATE-FILE THROW >R      \ Create object file, save name for xref
   'FILEIMAGE @ #FILEIMAGE @ R@ WRITE-FILE      \ Write file image to object file
   'FILEIMAGE @ FREE  R> CLOSE-FILE             \ Release buffer memory, close file, throw if write error
   2DROP THROW ;

: PROGRAM-SEALED ( -- )   -1 ?SEALED !  PROGRAM ;

END-PACKAGE
