{ ====================================================================
Catch & throw exception handling

Copyright 2000  FORTH, Inc.

This file implements Standard Forth exception handling.

Requires:  User variable CATCHER

Exports: CATCH  THROW

======================================================================
Overview of SwiftForth exception handling

KERNEL

The kernel implements CATCH and THROW, common to all platforms
(Windows, Linux, macOS):

SwiftForth\src\kernel\except.f (this file)

Each platform adds a handler to "trap" the system exceptions:

SwiftForth/src/kernel/linux/trap.f
SwiftForth/src/kernel/macos/trap.f
SwiftForth/src/kernel/win32/trap.f

Each of them registers its exception handler, named <EXCEPTION> with
the system. In Linux and macOS, this is done with multiple calls to
sigaction(), one for each "signal" were willing to handle. In
Windows, theres a single call to AddVectoredExceptionHandler() and
its up to the handler to determine if it will process the exception
or pass it along to be handled by another member of chain the of
registered handlers.

If the exception is for us, we receive from the OS the user context
for the exception. We capture the CPU registers from that context and
save it in the TRACEBACK buffer. We also save some return stack and
Forth data stack (as pointed to by ESP and EBP in the CPU register
context) for debug display.

We replace the EIP within the context with the address of THROWBACK
and return, indicating with our return status in EAX that we have
handled the exception and that the OS should resume execution from the
stored user context. The OS will then return back to the user-level
code using the stored context, which now will continue execution at
THROWBACK. And that just does IOR_EXCEPT THROW.

THROW will also capture return and data stacks in the same way if the
throw code passed to it is not IOR_EXCEPT. So no matter how you got a
THROW (system exception or user code calling THROW directly), you can
dump out the stacks for debug in a consistent way.

ELECTIVES

Up in the IDE world, we extend the systems .CATCH display to perform
.THROW, which will dump the captured CPU context and stacks if the
throw code is IOR_EXCEPT. All three platforms do this the same way in
this shared source file:

SwiftForth\src\ide\exception.f

The system-specific exception decode and mapping of the CPU context
registers is supplied in that platforms exception-map.f.

In Windows, we add the display of both kinds of debug info (THROW,
which shows stacks but not registers) and system exception (which
shows all of it) in a dialog box:

SwiftForth\src\ide\win32\exception-dialog.f

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Debug info

IOR_EXCEPT is the throw code for a system exception (a SIGNAL in MacOS
and Linux). The debug info that follows is captured by the system
exception handler before returning to its THROWBACK vector.

In the case where THROW is called directly (the ior is not
IOR_EXCEPT), THROW fills in the stack frame debug info.
-------------------------------------------------------------------- }

$DEADBEEF CONSTANT IOR_EXCEPT   \ Throw code for SIGNAL
VARIABLE EXCEPTION#             \ Exception code for debug handler

 20 CONSTANT #TRACE_REGS        \ Registers from ucontext record
256 CONSTANT #TRACE_RSTACK      \ Return stack trace cells
 20 CONSTANT #TRACE_DSTACK      \ Data stack trace cells

CREATE TRACEBACK
   #TRACE_REGS CELLS ALLOT      \ Registers
   #TRACE_RSTACK CELLS ALLOT    \ Return stack
   #TRACE_DSTACK CELLS ALLOT    \ Datastack

TRACEBACK #TRACE_REGS CELLS + EQU TRACE_RSTACK

{ --------------------------------------------------------------------
Exception frame operators

CATCH links a new exception frame on the return stack into the list
pointed to by user variable CATCHER, calls the execution token on the
stack.  If the called xt returns here, the exception frame is unlinked
and discarded, and 0 is returned on the data stack.

THROW takes a throw code n from the data stack.  If n is 0 or there is
no exception frame (user variable CATCHER contains 0), THROW simply
returns to its caller.  If n is non-zero, the exception is thrown via
the exception frame set up by the most recent call to CATCH.  The
return stack is set to the address held in CATCHER which is then
updated from the address on top of the return stack.  The parameter
stack pointer is also popped from the exception frame, the throw code
n is placed on the data stack, and control is returned immediately
after the call to CATCH.

THROWBACK is used by the system <EXCEPTION> handler. It returns
IOR_EXCEPT to indicate that a system exception occurred. Note that
THROW makes a copy of the stacks then the throw code is not
IOR_EXCEPT.
-------------------------------------------------------------------- }

28 EQU |FPENV|

CODE CATCH ( i*x xt -- j*x 0 | i*x n )
   |FPENV| # ESP SUB                    \ room for fpenv
   0 [ESP] FSTENV                       \ save fpenv
   0 [ESP] FLDCW                        \ restore control word
   'N [ESI] PUSH                        \ floating stack pointer
   'PERSONALITY [ESI] PUSH              \ personality
   'THIS [ESI] PUSH                     \ current class
   'SELF [ESI] PUSH                     \ current object
   'LF [ESI] PUSH                       \ current local variable frame
   'OB [ESI] PUSH                       \ current local objects frame
   EBP PUSH   ESI PUSH   EDI PUSH       \ Save Forth VM registers S, U, ORIGIN
   CATCHER [ESI] PUSH                   \ link to previous exception frame
   ESP CATCHER [ESI] MOV                \ CATCHER points to this frame
   EBX EAX MOV   POP(EBX)               \ Save xt, pop new TOS
   EDI EAX ADD   EAX CALL               \ Convert xt to addr, call it
   BEGIN ( *)                           \ Return point from call
   0 # PUSH   EAX POP                   \ Smudge former return addr
   CATCHER [ESI] POP                    \ Unlink frame
   |FPENV| 9 CELLS + # ESP ADD          \ Discard it
   PUSH(EBX)   EBX EBX XOR              \ Return 0
   RET   END-CODE

( *) CONSTANT 'CATCHING

CODE THROW ( k*x n -- k*x | i*x n )
   EBX EBX TEST   0= NOT IF             \ Skip if n=0
      IOR_EXCEPT # EBX CMP   0= NOT IF  \ Capture debug info if not IOR_EXCEPT
         ESI PUSH   EDI PUSH
         TRACE_RSTACK [EDI] EDI LEA
         ESP ESI MOV
         #TRACE_RSTACK # ECX MOV   REP MOVSD
         EBP ESI MOV
         #TRACE_DSTACK # ECX MOV   REP MOVSD
      EDI POP   ESI POP   THEN
      CATCHER [ESI] EAX MOV             \ Get pointer to exception frame
      EAX EAX TEST   0= NOT IF          \ Skip if no frame
         EAX ESP MOV   CATCHER [ESI] POP  \ Unlink frame
         0 # -CELL [ESP] MOV            \ Smudge former return addr
         EDI POP   ESI POP   EBP POP    \ Restore Forth VM registers
         'OB [ESI] POP                  \ local objects frame
         'LF [ESI] POP                  \ local variable frame
         'SELF [ESI] POP                \ current object
         'THIS [ESI] POP                \ current class
         'PERSONALITY [ESI] POP         \ personality
         'N [ESI] POP                   \ floating stack pointer
         FINIT                          \ floating stack init
         0 [ESP] FLDENV                 \ restore fpenv
         |FPENV| # ESP ADD              \ discarding its memory
         RET                            \ Return n
   THEN THEN
   POP(EBX)   RET   END-CODE            \ Bail out (n=0 or no frame) exits here

LABEL THROWBACK
   IOR_EXCEPT # EBX MOV              \ Return unique throw code
   ' THROW JMP   END-CODE
