{ ====================================================================
Interpreter support

Copyright 2001  FORTH, Inc.

This file supplies the Forth interpreter and supporting functions.
==================================================================== }

TARGET

{ ------------------------------------------------------------------------
Compilation

' <tick> looks up the next word in the dictionary and aborts if it is
not found. Returns xt.

[']  is the compile-time behaviour of <tick> .

POSTPONE appends the compilation semantics of "name" to the current
definition.

[COMPILE] is supplied for legacy support.  It unconditionally compiles
the next word in the input stream.  Its use is deprecated in favor of
POSTPONE.
------------------------------------------------------------------------ }

: HUH? ( flag -- )   0= ABORT" ? " ;

: ' ( "name" -- xt )
   BL WORD  DUP C@ HUH?  FIND HUH? ;

: ['] ( -- )
   ' POSTPONE LITERAL ;  IMMEDIATE

: POSTPONE ( "name" -- )
   BL WORD  DUP C@ HUH?  FIND  DUP HUH?  0< IF
      ( non-immediate)  POSTPONE LITERAL  POSTPONE COMPILE,
   ELSE  ( immediate)  COMPILE,  THEN ;  IMMEDIATE

: [COMPILE] ( "name" -- )   ' COMPILE, ;  IMMEDIATE

{ --------------------------------------------------------------------
Stack check
-------------------------------------------------------------------- }

: ?DSTACK ( -- )   DEPTH 0< -4 ?THROW ;

XDEFER ?STACK ( -- )   ' ?DSTACK ,

{ --------------------------------------------------------------------
Number
-------------------------------------------------------------------- }

| : NOLOCALS ( c-addr len -- 0 )   2DROP 0 ;

XDEFER (FINDEXT) ' NOLOCALS ,  ( c-addr len -- c-addr len 0 | xt flag )

: IS-KNOWN? ( c-addr len -- c-addr len 0 | xt flag )
   2DUP (FINDEXT) DUP IF 2NIP EXIT THEN DROP
   2DUP (FIND)    DUP IF 2NIP      THEN ;

: INTEGER? ( addr len 0 | ... xt -- addr len 0 | ... xt )
   DUP ?EXIT DROP
   2DUP ANY-NUMBER? DUP -EXIT
   1 = IF NIP NIP ['] LITERAL ELSE 2NIP ['] 2LITERAL THEN ;

( addr len 0 | ... xt -- addr len 0 | ... xt )
XDEFER IS-NUMBER?   ' INTEGER? ,

{ --------------------------------------------------------------------
Token handling

WHAT is the default behavior for NOT-DEFINED. It places a copy of the
undefined name in 'WHAT and does a -13 THROW ("undefined word").

NOT-DEFINED is extended later when the IDE electives are loaded.

WORD-COMPILE and WORD-INTERPRETER process tokens for compile and
interpret modes.
-------------------------------------------------------------------- }

CREATE 'WHAT  256 ALLOT
: WHAT ( c-addr len -- )   'WHAT PLACE  -13 THROW ;
XDEFER NOT-DEFINED ( a n -- )   ' WHAT ,

| : COMPILE ( xt flag -- )
   0< IF  COMPILE,  ELSE  EXECUTE  THEN ;

: WORD-COMPILER ( c-addr len -- )
   IS-KNOWN?  DUP IF  COMPILE EXIT  THEN
   IS-NUMBER? DUP IF  EXECUTE EXIT  THEN DROP
   NOT-DEFINED ;

: WORD-INTERPRETER ( c-addr len -- )
   IS-KNOWN? IF  EXECUTE EXIT  THEN
   0 IS-NUMBER? ?EXIT  NOT-DEFINED ;

{ --------------------------------------------------------------------
Interpreting and compiling

(INTERPRETING) returns true when input exausted or 0 when state is
wrong. STATE should be 0 while interpreting.

(COMPILING) is similar but while state is set.

The INTERPRET loop calls either COMPILING or INTERPRETING until they
return false (i.e. no more input or STATE has changed.
-------------------------------------------------------------------- }

: (INTERPRETING) ( -- flag )
   BEGIN
      STATE @ 0= WHILE
      BL WORD DUP C@ WHILE
      COUNT WORD-INTERPRETER ?STACK
   REPEAT 0<> ELSE 0 THEN ;

: (COMPILING) ( -- flag )
   BEGIN
      STATE @ WHILE
      BL WORD DUP C@ WHILE
      COUNT WORD-COMPILER ?STACK
   REPEAT 0<> ELSE 0 THEN ;

XDEFER COMPILING      ' (COMPILING) ,
XDEFER INTERPRETING   ' (INTERPRETING) ,

: INTERPRET ( -- )
   BEGIN
      STATE @ IF  COMPILING  ELSE  INTERPRETING  THEN
   UNTIL ;

{ --------------------------------------------------------------------
State manipulation
-------------------------------------------------------------------- }

: [ ( -- )   STATE OFF ;  IMMEDIATE
: ] ( -- )   STATE ON ;

{ ------------------------------------------------------------------------
Error reports and input prompting

(.CATCH) is the most primitive error reporting for catch/throw.
.CATCH vectors the error reporting routine. Default is (.CATCH).

(PROMPT) displays the ok prompt if not compiling.
PROMPT updates status and prompt. Default is (PROMPT).
------------------------------------------------------------------------ }

| : (.CATCH) ( n -- )
   DUP 1+ IF  DUP 2+ IF  ."  Catch = " DUP .
   ELSE  @ERRMSG TYPE  THEN THEN DROP ;

XDEFER .CATCH   ' (.CATCH) ,

| : (PROMPT)
   STATE @ 0= IF  ."  ok" THEN CR ;

XDEFER PROMPT   ' (PROMPT) ,

{ ------------------------------------------------------------------------
Quit loop

(QUIT) is the heart of the text interpreter: updating status,
refilling the input stream, interpreting the input stream, and
prompting the user for more -- over and over and over again.

QUIT is the main loop for the entire system. It clears the return
stack, CATCHES (QUIT), and reports errors.

INTERPRET is called via CATCH, so all errors should come back here.
The data stack is cleared on any error.
------------------------------------------------------------------------ }

XDEFER /NDP    0 ,

: (QUIT) ( -- )
   BEGIN  REFILL DROP  INTERPRET PROMPT  AGAIN ;

: /INTERPRETER ( -- )
   'TIB CELL+ @ 'TIB !  'SOURCE-ID OFF  BLK OFF   POSTPONE [ ;

: QUIT ( -- )
   BEGIN
      R0 @ RP!         \ clear return stack
      /INTERPRETER
      ['] (QUIT) CATCH .CATCH CR
      S0 @ SP!  /NDP
   AGAIN ;

{ ------------------------------------------------------------------------
Input stream preservation

Save and restore information about the current input stream. This
should be sufficient to allow EVALUATE, INCLUDE, and LOAD to be nested
in arbitrary orders. No assumption can be made about the contents or
order of parameters returned by SAVE-INPUT.
------------------------------------------------------------------------ }

: RESTORE-INPUT ( ... n -- 0 )   8 <> ABORT" Bad arguments to RESTORE-INPUT"
   >MAPPED 2!  LINE !  BLK !  'SOURCE-ID !  >IN !  #TIB 2!  0 ;

: SAVE-INPUT ( -- ... n )
   #TIB 2@  >IN @  'SOURCE-ID @  BLK @  LINE @  >MAPPED 2@  8 ;

{ ------------------------------------------------------------------------
Evaluate

Given a string and length on the stack, evaluate it in the current
system context.
------------------------------------------------------------------------ }

: .ERRLINE ( -- )   'TIB @ >IN @ 2DUP  BRIGHT TYPE NORMAL
   #TIB @ SWAP /STRING TYPE  SPACE ;

: EVALUATE ( i*x c-addr u -- j*x )
   SAVE-INPUT N>R
   ( c-addr u) #TIB 2!  >IN OFF  LINE OFF  BLK OFF  -1 'SOURCE-ID !
   ['] INTERPRET CATCH ( *)
   ( *) DUP IF  CR .ERRLINE  POSTPONE [   THEN
   NR> RESTORE-INPUT DROP ( *) THROW ;

{ ------------------------------------------------------------------------
Comments

\  comments to end of line.

FILE-SKIP  ignores data from the current file input stream until a
matching character is found, refilling as necessary.

TEXT-SKIP  ignores data from the current input stream until a matching
character is found.

(  comments to the next ) . {  comments to the next <right curly
brace> . .(  prints text until the next ) .

------------------------------------------------------------------------ }

: \\ ( -- )
   SOURCE-ID IF  BEGIN REFILL 0= UNTIL  THEN ;

: \ ( " ccc")
   BLK @ IF  >IN @  64 + -64 AND  ELSE
   #TIB @  THEN  >IN ! ;  IMMEDIATE

: SKIP-PAST ( char -- )
   BEGIN
      /SOURCE THIRD SCAN NIP IF
         PARSE 2DROP EXIT
      THEN
      REFILL 0=
   UNTIL DROP ;

: (   [CHAR] ) SKIP-PAST ; IMMEDIATE
: {   [CHAR] } SKIP-PAST ; IMMEDIATE

: ECHO-UNTIL ( char -- )
   BEGIN
      /SOURCE THIRD SCAN NIP IF
         PARSE TYPE EXIT
      THEN
      /SOURCE TYPE CR
      REFILL 0=
   UNTIL DROP ;

: .( ( -- )    [CHAR] ) ECHO-UNTIL ; IMMEDIATE
: ?( ( -- )   CR .( ; IMMEDIATE

{ ------------------------------------------------------------------------
More defining word behaviour

(;CODE) fixes the compiled CALL of the most recent word so that it
calls the code following where it was called from, normally in a
defining word.

DOES> defines a new behavior for word defined by CREATE.

RECURSE forces a word to call itself.

ASSIGN writes a pseudo-xt for the code following it into a variable.
For example:

        VARIABLE 'DOTHIS

        : HEREITIS
           'DOTHIS ASSIGN HERE 100 DUMP ;

        HEREITIS

        'DOTHIS @EXECUTE

   ...will execute the code following the ASSIGN in HEREITIS.
------------------------------------------------------------------------ }

: (;CODE) ( -- )
   R> ( addr to call to )
   LAST @ NAME> >BODY -
   LAST @ NAME> >CODE $E8 OVER C! 1+ ! ;

: DOES> ( -- )
   POSTPONE (;CODE)
   POSTPONE (DOES>) ;  IMMEDIATE

: RECURSE ( -- )
   LAST CELL+ CELL+ @ CODE> COMPILE, ;  IMMEDIATE

: ASSIGN  ( addr -- )
   R> CODE> SWAP ! ;
