{ ====================================================================
System console

Copyright 2007  FORTH, Inc.

This driver provides console I/O using STDIN and STDOUT for input and
output, respectively.  Output can also be directed to STDERR.
==================================================================== }

TARGET

{ --------------------------------------------------------------------
Initialization and exit

FD-IN and FD-OUT have headers, so they can be read and set by user.

INIT-TERM preserves the current tty attributes and initializes ?TTY.
RESET-TERM restores tty attributes.

SET-TERM configures the tty for key input.  (Also used later by
SIGCONT signal handler callback.)

[CC] defines fields in the TS0 termios structure.
>STDIN sets console input from STDIN (default).
>STDOUT directs console output to STDOUT (default).
>STDERR directs console output to STDERR.
-------------------------------------------------------------------- }

0 CONSTANT STDIN                \ fid of STDIN
1 CONSTANT STDOUT               \ fid of STDOUT
2 CONSTANT STDERR               \ fid of STDERR

VARIABLE ?TTY                   \ flag set if console is a tty ***DO NOT CHANGE THIS***

\ The size and format of the termios structure
\ is different between macOS and Linux.

| CREATE TS0  48 ALLOT          \ STDIN get attributes (original)
| CREATE TS1  48 ALLOT          \ STDIN set attributes (no echo, for KEY input)
CREATE FD-IN   STDIN ,          \ File descriptor of input device
CREATE FD-OUT  STDOUT ,         \ File descriptor of output device

: >STDIN ( -- )    STDIN  FD-IN  ! ;
: >STDOUT ( -- )   STDOUT FD-OUT ! ;
: >STDERR ( -- )   STDERR FD-OUT ! ;

INTERPRETER

: [CC] ( i -- )   CREATE 16 + C,
   DOES> ( -- uchar )  C@ TS0 + C@ ;

TARGET

\ Different from Linux

 0 [CC] VEOF            \ End-of-file (Ctrl-D)
 1 [CC] VEOL            \ Additional EOL char (NUL)

 3 [CC] VERASE          \ Erase single character (Del, BS)
 4 [CC] VWERASE         \ Word erase (Ctrl-W)
 5 [CC] VKILL           \ Kill line (Ctrl-U, Ctrl-X)
 6 [CC] VREPRINT        \ Reprint unread characters (Ctrl-R)

 8 [CC] VINTR           \ Interrupt, send SIGINT (Ctrl-C)
 9 [CC] VQUIT           \ Quit, send SIGQUIT (FS, Ctrl-\)
10 [CC] VSUSP           \ Suspend, send SIGTSTP (Ctrl-Z)
12 [CC] VSTART          \ Restart output stopped by VSTOP (Ctrl-Q)
13 [CC] VSTOP           \ Stop output until VSTART (Ctrl-S)
15 [CC] VDISCARD        \ Toggle: start/stop discarding pending output (Ctrl-O)
16 [CC] VMIN            \ Min chars for non-canonical read
17 [CC] VTIME           \ Time in 0.1 sec units for non-canonical read

-HEADS

HEX

CREATE &ATTR
   FFFF9000 , 0 ,    \ c_iflag (see termios.txt)
   FFFFFFFC , 3 ,    \ c_oflag
   FFFF00FF , 4B00 , \ c_cflag
   FFFFFA00 , 4C7 ,  \ c_lflag

HEX

\ Some ioctl command codes...

40087468 CONSTANT TIOCGSIZE
0 CONSTANT TCSANOW

DECIMAL

: INIT-TERM ( -- )
   STDIN TS0 tcgetattr 0= ?TTY !                \ Get entry settings, set ?TTY if return code is okay
   TS0 TS1 48 CMOVE                             \ Keep original in TS0 for reset
   &ATTR  16 0 DO  DUP 2@  TS1 I + @ AND OR     \ Apply masks
   TS1 I + !  8 +  4 +LOOP DROP ;               \ ...working copy in TS1

+HEADS

: RESET-TERM ( -- )                             \ Reset terminal back to TS0
   STDIN TCSANOW TS0 tcsetattr DROP ;

: SET-TERM ( -- )                               \ Set terminal input from result in TS1
   STDIN TCSANOW TS1 tcsetattr DROP ;

{ --------------------------------------------------------------------
Terminate

EXITSTATUS is the exit status returned to the OS.

_BYE terminates SwiftForth with the code in EXITSTATUS and is the
default action for BYE which is extended later.
-------------------------------------------------------------------- }

VARIABLE EXITSTATUS

: _BYE ( -- )
   CR  ?TTY @ IF  RESET-TERM  NORMAL  THEN
   EXITSTATUS @ sys_exit ;

: .BYE ( addr u -- )
   STDERR -ROT write DROP  _BYE ;

XDEFER BYE   ' _BYE ,

-HEADS

{ --------------------------------------------------------------------
Console I/O

(C-KEY?) tests the console input device.
(C-KEY) waits for and returns a keystroke from the console.
(C-TYPE) writes a string to the console.
(C-EMIT) writes a character to the console.
(C-GETSIZE) returns the current console window dimensions.
-------------------------------------------------------------------- }

VARIABLE CBUF

CREATE POLLFD
   ( stdin) 0 ,  ( POLLIN) 1 W,  ( result) 0 W,

: (C-KEY?) ( -- flag )
   POLLFD FD-IN @ OVER ! 1 0 poll NEGATE ;

: (C-KEY) ( -- char )
   FD-IN @ CBUF 1 read DROP  CBUF C@ ;

: (C-TYPE) ( addr n -- )
   FD-OUT @ -ROT write DROP ;

: (C-EMIT) ( char -- )
   CBUF C!  FD-OUT @ CBUF 1 write DROP ;

: (C-GETSIZE) ( -- x y )   R-BUF        \ temporary place for struct winsize
   FD-IN @ TIOCGSIZE R@ ioctl DROP      \ read winsize
   R> @ HILO ;                          \ return x,y size

{ ---------------------------------------------------------------------
ANSI/VT100 terminal vectored behaviors

(C-GETXY) returns 0 0 if the response is not the position escape
sequence from the terminal.

(C-PAGE) clears entire screen buffer.
--------------------------------------------------------------------- }

: (C-CR) ( -- )   $0A (C-EMIT) ;

: CSI ( -- )   27 EMIT  ." [" ;

: (C-ATXY) ( x y -- )   BASE @ >R DECIMAL  CSI
   1+ 0 U.R  ." ;"  1+ 0 U.R  ." H"  R> BASE ! ;

: (GETPOS) ( -- n )
   0  BEGIN  KEY  DUP 48 58 WITHIN WHILE  48 - SWAP 10 * +
   REPEAT DROP  ?DUP IF  1-  THEN ;

: (C-GETXY) ( -- x y )
   0 0  CSI ." 6n"  BEGIN  KEY 27 =  UNTIL
   KEY [CHAR] [ = IF  2DROP  (GETPOS) (GETPOS) SWAP  THEN ;

: (C-PAGE) ( -- )   CSI ." H"  CSI  ." 2J" ;

CREATE 'ATTRIB
   0 C,         \ Normal
   7 C,         \ Inverse
  31 C,         \ Bright
  36 C,         \ Bold
   8 C,         \ Invisible

: (C-ATTRIBUTE) ( n -- )
   BASE @ >R DECIMAL  0 MAX 4 MIN 'ATTRIB + C@  CSI ." 0"
   ?DUP IF  ." ;"  DUP (.) TYPE 8 > IF ." ;1"  THEN THEN
   ." m"  R> BASE ! ;

{ --------------------------------------------------------------------
Non-tty input
-------------------------------------------------------------------- }

: (N-XY) ( -- x y )   0 0 ;

: (N-ACCEPT) ( c-addr u1 -- u2 )
   OVER SWAP 0 DO  STDIN OVER 1 read IF
      DUP C@ $0A = IF  LEAVE  THEN  1+
      ELSE  I 0= IF BYE THEN  LEAVE
   THEN LOOP  SWAP - ;

: (N-KEY) ( -- char )
   STDIN CBUF 1 read 0= IF  BYE  THEN  CBUF C@ ;

{ --------------------------------------------------------------------
TTY-MODE is a personality vector for the interactive console.

NOTTY-MODE is a personality vector for a non-interactive SwiftForth
session (stdin not a tty; e.g.: "echo pad 20 dump | sf".

/CONSOLE sets the correct personality and performs any necessary
initialization.
-------------------------------------------------------------------- }

CREATE TTY-MODE
        16 ,            \ datasize
        19 ,            \ maxvector
         0 ,            \ PHANDLE
         0 ,            \ PREVIOUS
   ' NOOP ,             \ INVOKE    ( -- )
   ' NOOP ,             \ REVOKE    ( -- )
   ' NOOP ,             \ /INPUT    ( -- )
   ' (C-EMIT) ,         \ EMIT      ( char -- )
   ' (C-TYPE) ,         \ TYPE      ( addr len -- )
   ' (C-TYPE) ,         \ ?TYPE     ( addr len -- )
   ' (C-CR) ,           \ CR        ( -- )
   ' (C-PAGE) ,         \ PAGE      ( -- )
   ' (C-ATTRIBUTE) ,    \ ATTRIBUTE ( n -- )
   ' (C-KEY) ,          \ KEY       ( -- char )
   ' (C-KEY?) ,         \ KEY?      ( -- flag )
   ' (C-KEY) ,          \ EKEY      ( -- echar )
   ' (C-KEY?) ,         \ EKEY?     ( -- flag )
   ' (C-KEY) ,          \ AKEY      ( -- char )
   ' 2DROP ,            \ PUSHTEXT  ( addr len -- )
   ' (C-ATXY) ,         \ AT-XY     ( x y -- )
   ' (C-GETXY) ,        \ GET-XY    ( -- x y )
   ' (C-GETSIZE) ,      \ GET-SIZE  ( -- x y )
   ' (ACCEPT) ,       	\ ACCEPT    ( addr u1 -- u2 )

CREATE NOTTY-MODE
        16 ,            \ datasize
        19 ,            \ maxvector
         0 ,            \ PHANDLE
         0 ,            \ PREVIOUS
   ' NOOP ,             \ INVOKE    ( -- )
   ' NOOP ,             \ REVOKE    ( -- )
   ' NOOP ,             \ /INPUT    ( -- )
   ' (C-EMIT) ,         \ EMIT      ( char -- )
   ' (C-TYPE) ,         \ TYPE      ( addr len -- )
   ' (C-TYPE) ,         \ ?TYPE     ( addr len -- )
   ' (C-CR) ,           \ CR        ( -- )
   ' NOOP ,             \ PAGE      ( -- )
   ' DROP ,             \ ATTRIBUTE ( n -- )
   ' (N-KEY) ,          \ KEY       ( -- char )
   ' (C-KEY?) ,         \ KEY?      ( -- flag )
   ' (N-KEY) ,          \ EKEY      ( -- echar )
   ' (C-KEY?) ,         \ EKEY?     ( -- flag )
   ' (N-KEY) ,          \ AKEY      ( -- char )
   ' 2DROP ,            \ PUSHTEXT  ( addr len -- )
   ' 2DROP ,            \ AT-XY     ( x y -- )
   ' (N-XY) ,           \ GET-XY    ( -- x y )
   ' (N-XY) ,           \ GET-SIZE  ( -- x y )
   ' (N-ACCEPT) ,       \ ACCEPT    ( addr u1 -- u2 )

+HEADS

: /CONSOLE ( -- )   NOTTY-MODE  INIT-TERM
   ?TTY @ IF  DROP TTY-MODE  SET-TERM  THEN
   'PERSONALITY ! ;
