{ ====================================================================
SwiftForth nucleus

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Origin Locator

When the program starts, the base address of the SwiftForth memory
image is loaded in EDI.

This is aligned just after the headers in the image file.  It makes no
difference to the application running, but makes looking at a binary
file simpler.

ORIGIN returns address of system origin pointed to by EDI.
+ORIGIN and -ORIGIN add or subtract against the origin.
-------------------------------------------------------------------- }

ICODE ORIGIN ( -- addr )
   PUSH(EBX)   EDI EBX MOV   RET   END-CODE

ICODE +ORIGIN ( n -- addr )
   EDI EBX ADD   RET   END-CODE

ICODE -ORIGIN ( addr -- n )
   EDI EBX SUB   RET   END-CODE

{ ------------------------------------------------------------------------
Runtime defining words behaviour

Each root class of defining word gets its own runtime subroutine.
These routines are called by any instance of the corresponding
defining word. The call is made because of the Forth model, which
yields the address of a given bit of data not constant.

Note that VALUE has its own behaviour even though it is the same as
CONSTANT.  This is so the "TO" family can validate behaviour.
------------------------------------------------------------------------ }

CODE (CREATE) ( -- addr )
   PUSH(EBX)   EBX POP   RET   END-CODE

CODE (CONSTANT) ( -- x )
   EAX POP   PUSH(EBX)   0 [EAX] EBX MOV
   RET   END-CODE

CODE (2CONSTANT) ( -- d )
   -2 CELLS [EBP] EBP LEA   EBX CELL [EBP] MOV
   EBX POP   CELL [EBX] EAX MOV
   EAX 0 [EBP] MOV   0 [EBX] EBX MOV
   RET   END-CODE

CODE (VALUE) ( -- x )
   EAX POP   PUSH(EBX)   0 [EAX] EBX MOV
   RET   END-CODE

CODE (2VALUE) ( -- x1 x2 )
   -2 CELLS [EBP] EBP LEA   EBX CELL [EBP] MOV
   EBX POP   CELL [EBX] EAX MOV
   EAX 0 [EBP] MOV   0 [EBX] EBX MOV
   RET   END-CODE

CODE (USER)
   EAX POP   PUSH(EBX)   ESI EBX MOV
   0 [EAX] EBX ADD   RET   END-CODE

ICODE (LITERAL)  ( -- n )
   PUSH(EBX)   0 # EBX MOV      \ This immediate value is filled in by LITERAL
   RET   END-CODE

CODE (DEFER)
   ECX POP   0 [ECX] ECX MOV   ECX ECX OR
   0<> IF   EDI ECX ADD   ECX JMP   THEN
   RET   END-CODE

{ ------------------------------------------------------------------------
DOES behaviour

On its entry, the top item on the processor stack will point to the
body of the defined word, EBX will be top of data stack, and the rest
of the data stack will be in memory at EBP
------------------------------------------------------------------------ }

ICODE (DOES>)
   PUSH(EBX)   EBX POP   RET   END-CODE

{ ------------------------------------------------------------------------
Branch and Loop Primitives

The branching instructions all end with a 32-bit unresolved branch (in
both 32- and 64-bit implementations), which is patched with the
correct offset by their compiling words. These words are never called
directly; they're always be inlined by their compiling words.

(IF) is a conditional branch, which takes the branch if the top of
stack is non-zero. (ELSE) is an unconditional branch. These two are
sufficient to build the non-iterated loop structures.

(LOOP) and (+LOOP) are the iterated loop terminators. (LOOP) uses a
fixed increment of 1, while (+LOOP) uses an arbitrary increment. The
indices are kept on the return stack, and are offset by half the
number circle so that a full-range loop can be implemented. Loop
termination is arranged to be on overflow.

(DO) pushes the prepared indices onto the return stack. (?DO) only
executes if the start and end indices are different.

UNLOOP provides a mechanism for discarding the DO..LOOP indices. This
is useful if one desires to exit from a DO..LOOP immediately.

I J and K calculate the correct DO..LOOP index from the parameters on
the return stack.
------------------------------------------------------------------------ }

ICODE (ELSE)
   HERE $400 + JMP   RET   END-CODE

ICODE (IF)
   EBX EBX OR   0 [EBP] EBX MOV
   CELL [EBP] EBP LEA   HERE $400 + JZ
   RET   END-CODE

ICODE (OF) ( x1 x2 -- x1 x2 | x1 )
   0 [EBP] EBX CMP  0<> IF
      0 [EBP] EBX MOV   CELL [EBP] EBP LEA
   HERE $400 + JMP   THEN   RET   END-CODE

CODE (DO) ( x1 x2 -- )
   EDX POP
   0 [EBP] EAX MOV
   $80000000 # EAX ADD
   EAX PUSH
   EAX EBX SUB
   EBX PUSH
   CELL [EBP] EBX MOV
   2 CELLS [EBP] EBP LEA
   EDX JMP
   RET   END-CODE

ICODE (?DO) ( x1 x2 -- x1 x2 )
   0 [EBP] EBX CMP
   0= IF
      CELL [EBP] EBX MOV
      2 CELLS [EBP] EBP LEA
      HERE $400 + JMP
   THEN  RET   END-CODE

ICODE (LOOP) ( -- )
   0 [ESP] INC
   HERE $400 + JNO
   RET   END-CODE

ICODE (+LOOP) ( n -- )
   EBX 0 [ESP] ADD
   0 [EBP] EBX MOV
   CELL [EBP] EBP LEA
   HERE $400 + JNO
   RET   END-CODE

{ ------------------------------------------------------------------------
Execute and Exit

In this native code implementation, SwiftForth uses the processor's
return stack for its return stack. An EXIT operation is exactly
equivalent to the processor RET instruction.

EXIT is coded with an extra RET because it is designed to be compiled
inline, and the inline compiler will discard the final RET opcode.

?EXIT tests the top of stack and returns if non-zero. -EXIT tests and
returns if zero.

?IOR will exit with an ior intact, or will continue with a zero
dropped. just like: "?DUP ?EXIT"

EXECUTE works only with an XT. This is an ANS-Forth concept, which is
not necessarily related to an address in memory. In this
implementation, an XT is an offset from the beginning of the kernel.

EXECUTE and @EXECUTE must not be inlined.

Neither EXECUTE nor @EXECUTE will operate on a zero. This is
convenient for execution vectors that are uninitialized.
------------------------------------------------------------------------ }

ICODE EXIT ( -- )
   RET                              \ this return fools the inliner
   RET   END-CODE

ICODE ?EXIT ( flag -- )
   EBX EBX OR
   0 [EBP] EBX MOV
   CELL [EBP] EBP LEA
   0<> IF RET THEN
   RET   END-CODE

ICODE ?IOR ( ior -- ior )  ( 0 -- )
   EBX EBX OR
   0<> IF RET THEN
   0 [EBP] EBX MOV
   CELL [EBP] EBP LEA
   RET   END-CODE

ICODE -EXIT ( flag -- )
   EBX EBX OR
   0 [EBP] EBX MOV
   CELL [EBP] EBP LEA
   0= IF RET THEN
   RET   END-CODE

CODE EXECUTE ( xt -- )
   EBX ECX MOV                      \ save xt
   POP(EBX)                         \ refresh tos
   ECX ECX OR  0<> IF
      EDI ECX ADD                   \ so we have an image address
      ECX JMP                       \ and jump to it
   THEN                             \ otherwise, fall thru and go on
   RET   END-CODE

CODE @EXECUTE ( addr -- )
   0 [EBX] ECX MOV                  \ read an xt from the address
   POP(EBX)                         \ refresh tos
   ECX ECX OR  0<> IF
      EDI ECX ADD                   \ so we have an image address
      ECX JMP                       \ and jump to it
   THEN                             \ otherwise, fall thru and go on
   RET   END-CODE

ICODE CALL ( addr -- )
   EBX EAX MOV
   POP(EBX)
   EAX CALL
   RET   END-CODE

{ ------------------------------------------------------------------------
Internal Register Manipulation

The system uses the following internal registers:

   ESI = User area pointer
   EDI = System origin
   EBP = Data stack pointer (tos in EBX)
   ESP = Return stack pointer

The following words set and read these registers from Forth. Note that
arbitrary values written to any of these registers will cause a system
crash, particularly EBP.
------------------------------------------------------------------------ }

ICODE UP@ ( -- a )
   PUSH(EBX)                        \ save tos
   ESI EBX MOV                      \ get user area pointer
   RET   END-CODE

ICODE RP@ ( -- a )
   PUSH(EBX)                        \ save tos
   ESP EBX MOV                      \ get return stack pointer
   RET   END-CODE

ICODE SP@ ( -- a )
   PUSH(EBX)                        \ save tos
   EBP EBX MOV                      \ get stack pointer
   RET   END-CODE

ICODE UP! ( a -- )
   EBX ESI MOV                      \ set user area pointer
   POP(EBX)                         \ refresh tos
   RET   END-CODE

ICODE RP! ( a -- )
   EBX ESP MOV                      \ set return stack pointer
   POP(EBX)                         \ refresh tos
   RET   END-CODE

ICODE SP! ( a -- )
   EBX EBP MOV                      \ set stack pointer
   POP(EBX)                         \ refresh tos
   RET   END-CODE
