{ ====================================================================
String manipulation

Copyright 2001  FORTH, Inc.
Rick VanNorman

==================================================================== }

TARGET

{ --------------------------------------------------------------------
Strings are represented on the stack in FORTH in three forms:

1) the address of the count byte, limited to 255 characters,
2) an address and a count,
3) an address of a zero-terminated string.

COUNT takes the address of a count byte and returns the address of the
string and the count byte on the stack.

CMOVE moves a string given its origin, destination, and length.  The
first byte is moved first, and if the memory areas overlap with the
destination above the origin, data destruction is possible.

CMOVE> also moves a string, but moves the last byte first.
-------------------------------------------------------------------- }

ICODE COUNT ( a -- a+1 [a] )
   EAX EAX XOR   0 [EBX] AL MOV  EBX INC
   PUSH(EBX)   EAX EBX MOV
   RET END-CODE

CODE CMOVE ( from to count -- )
   EBX EBX TEST   0= NOT IF
      0 [EBP] EDX MOV   CELL [EBP] ECX MOV
      BEGIN   0 [ECX] AL MOV   AL 0 [EDX] MOV
      ECX INC   EDX INC   EBX DEC   0= UNTIL
   THEN   2 CELLS [EBP] EBX MOV   3 CELLS [EBP] EBP LEA
   RET   END-CODE

CODE CMOVE> ( from to count -- )
   EBX EBX TEST   0= NOT IF
      0 [EBP] EDX MOV   CELL [EBP] ECX MOV
      BEGIN   -1 [EBX] [ECX] AL MOV
         AL -1 [EBX] [EDX] MOV
   EBX DEC   0= UNTIL   THEN
   2 CELLS [EBP] EBX MOV   3 CELLS [EBP] EBP LEA
   RET   END-CODE

{ ------------------------------------------------------------------------
MOVE determines which of CMOVE and CMOVE> to use to guarantee
   that there is no overlap.

FILL memory with an arbitrary byte.
ERASE fills memory with zeros.
BLANK fills memory with blanks.
------------------------------------------------------------------------ }

: MOVE ( from to count -- )
   >R  2DUP U< IF  R> CMOVE>  ELSE  R> CMOVE  THEN ;

CODE FILL ( c-addr u char -- )
   0 [EBP] ECX MOV   ECX ECX TEST       \ Get count, skip if 0
   0= NOT IF   CELL [EBP] EDX MOV          \ Get string addr
      BEGIN   BL 0 [EDX] MOV            \ Write char
      EDX INC   ECX DEC   0= UNTIL      \ Increment pointer, loop
   THEN   2 CELLS [EBP] EBX MOV               \ Drop parameters
   3 CELLS [EBP] EBP LEA
   RET   END-CODE

: ERASE ( c-addr u -- )
   0 FILL ;                         \ fill with zeros

: BLANK ( c-addr u -- )
   32 FILL ;                        \ fill with blanks

{ --------------------------------------------------------------------
Adjust string

/STRING trims a string's leading characters and returns the new
   string.

STRING/ discards all but the last _n_ characters of the string.
-------------------------------------------------------------------- }

ICODE /STRING ( a u n -- a+n u-n )
   EBX CELL [EBP] ADD   0 [EBP] EBX SUB
   CELL [EBP] EBP LEA   EBX NEG
   RET   END-CODE

: STRING/ ( a n n -- a n )
   OVER SWAP - 0 MAX  /STRING ;

{ ------------------------------------------------------------------------
String scanners

SKIP advances past leading characters in the specified string.

SCAN looks for a particular character.

-TRAILING removes the blanks from the end of a string by changing
   the count of the string.
------------------------------------------------------------------------ }

CODE SKIP ( c-addr1 u char -- c-addr2 u2 )
   EDI PUSH   EBX EAX MOV
   0 [EBP] ECX MOV   ECXNZ IF
      CELL [EBP] EDI MOV   REPE SCASB
      0<> IF   ECX INC   EDI DEC   THEN
   EDI CELL [EBP] MOV   THEN
   ECX EBX MOV   CELL [EBP] EBP LEA
   EDI POP   RET   END-CODE

CODE SCAN ( c-addr1 u char -- c-addr2 u2 )
   EDI PUSH   EBX EAX MOV
   0 [EBP] ECX MOV   ECXNZ IF
      CELL [EBP] EDI MOV   REPNE SCASB
      0= IF   ECX INC   EDI DEC   THEN
   EDI CELL [EBP] MOV   THEN
   ECX EBX MOV   CELL [EBP] EBP LEA
   EDI POP   RET   END-CODE

CODE -TRAILING ( c-addr u1 -- c-addr u2 )
   EBX EBX TEST   0= NOT IF
      0 [EBP] ECX MOV   BEGIN
         32 # -1 [EBX] [ECX] BYTE CMP
      0= WHILE   EBX DEC   0= UNTIL
   THEN THEN   RET   END-CODE

{ --------------------------------------------------------------------
String building functions

APPEND adds a one string to an existing string.
PLACE puts a string into a buffer.
DIVIDE splits a string at a position, returning left and right parts.

SPLIT divides a string at a given character. The first part of the
string is on top, the remaining part underneath. The remaining part
begins with the scanned-for character.
-------------------------------------------------------------------- }

: APPEND ( from len to -- )
   2DUP 2>R  COUNT + SWAP MOVE  2R> C+! ;

: PLACE ( from n to -- )
   2DUP 2>R  1+ SWAP MOVE  2R> C! ;

: DIVIDE ( c-addr len n -- c-addr n c-addr+n len-n )
   >R  2DUP R@ /STRING 0 MAX  2SWAP R> MIN ;

: SPLIT ( addr len char -- addr len addr len )
   >R 2DUP R> SCAN ROT OVER - >R ROT R> ;

{ --------------------------------------------------------------------
Null-terminated strings

These string operators support null-terminated (ASCIIZ) strings, which
are common in the Windows, Linux, and macOS environments.

ZCOUNT measures an ASCIIZ string, max 65535 length to prevent
runaways, and returns an address and length, suitable for TYPE.

ZLENGTH returns the length of an ASCIIZ string.

ZPLACE puts a string at an address and zero terminates it.

ZAPPEND appends a string to an ASCIIZ string.
------------------------------------------------------------------------ }

: ZCOUNT ( zaddr -- addr n )
   DUP DUP IF  65535 0 SCAN DROP OVER - THEN ;

: ZLENGTH ( zaddr -- n )
   ZCOUNT NIP ;

: ZPLACE ( from n to -- )
   TUCK OVER + >R  CMOVE  0 R> C! ;

: ZAPPEND ( from n to -- )
   ZCOUNT + ZPLACE ;

{ ------------------------------------------------------------------------
String runtime behaviours

There are three basic compiled string behaviours in SwiftForth.  All
strings are compiled alike, with a count byte and a trailing NULL.

(S") returns address and count of the string.
(C") returns the address of the count byte of the string.
(Z") returns the address of the first byte of the string.
------------------------------------------------------------------------ }

: (S") ( -- c-addr u )   R> COUNT  2DUP + 1+ >R ;
: (C") ( -- c-addr )   R> DUP   COUNT + 1+ >R ;
: (Z") ( -- c-addr )   R> COUNT  OVER + 1+ >R ;

{ ------------------------------------------------------------------------
String comparison

Compare the string specified by c-addr1 u1 to the string specified by
c-addr2 u2. The strings are compared beginning at the given addresses,
character by character, up to the length of the shorter string or
until a difference is found. If the two strings are identical, n is 0.
If the two strings are identical up to the length of the shorter
string, n is -1 if u1 is less than u2 and 1 otherwise. If the two
strings are not identical up to the length of the shorter string, n is
-1 if the first non-matching character in the string specified by
c-addr1 u1 has a lesser numeric value than the corresponding character
in the string specified by c-addr2 u2 and 1 otherwise.

REPE-CMPSB(UC) is a subroutine which replaces REPE CMPSB with the
distinction of being case insensitive. No free registers.

COMPARE tests the two strings for equality, returning a multi-flag. -1
if string 1 is less than string 2, +1 if greater, 0 if equal.

COMPARE(NC) is the same, but non-case-sensitive.

COMPARE(CS) is a string compare that respects the system variable CAPS.

-MATCH is a polyForth primitive that looks for (s #) in (a n), and
returns the address after the match if found and false or (if not
found) garbage and true.
------------------------------------------------------------------------ }

LABEL REPE-CMPSB(UC)
   EAX PUSH   ECXNZ IF
      BEGIN ( *)
         0 [EDI] AL MOV  EDI INC
         CHAR a # AL CMP   U>= IF
            CHAR z # AL CMP   U<= IF
               32 # AL XOR
         THEN THEN
         0 [ESI] AH MOV   ESI INC
         CHAR a # AH CMP   U>= IF
            CHAR z # AH CMP   U<= IF
               32 # AH XOR
         THEN THEN
         AL AH CMP   0= WHILE   ( *) LOOPE
   THEN THEN
   EAX POP   RET   END-CODE

CODE COMPARE ( c-addr1 len1 c-addr2 len2 -- n )
   EBX ECX MOV   EBX EBX SUB
   CELL [EBP] ECX CMP   0<> IF
      0< IF   EBX INC
         ELSE   EBX DEC   CELL [EBP] ECX MOV
   THEN THEN
   ECXNZ IF   ESI PUSH   EDI PUSH
      2 CELLS [EBP] ESI MOV   0 [EBP] EDI MOV
      REPE CMPSB   0<> IF
         CC IF   1 # EBX MOV   ELSE   -1 # EBX MOV
      THEN THEN
   EDI POP   ESI POP   THEN
   3 CELLS [EBP] EBP LEA
   RET   END-CODE

CODE COMPARE(NC) ( c-addr1 len1 c-addr2 len2 -- n )
   EBX ECX MOV   EBX EBX SUB
   CELL [EBP] ECX CMP   0<> IF
      CS IF   EBX INC
      ELSE   EBX DEC   CELL [EBP] ECX MOV
   THEN THEN
   ECXNZ IF
      ESI PUSH   EDI PUSH
      2 CELLS [EBP] ESI MOV   0 [EBP] EDI MOV
      REPE-CMPSB(UC) CALL   0<> IF
         CC IF   1 # EBX MOV
            ELSE   -1 # EBX MOV
      THEN THEN
   EDI POP   ESI POP   THEN
   3 CELLS [EBP] EBP LEA
   RET   END-CODE

: COMPARE(CS) ( addr1 u1 addr2 u2 -- n )        \ Use current case-sensitivity
   CAPS @ IF  COMPARE(NC)  ELSE  COMPARE  THEN ;

CODE -MATCH ( a n s # -- a t)
   ESI PUSH   EDI PUSH   EBX EDX MOV
   0 [EBP] EAX MOV   CELL [EBP] ECX MOV
   2 CELLS [EBP] EDI MOV   EBX EAX XCHG
   2 CELLS [EBP] EBP LEA   0 [EBX] AL MOV
   EDX DEC   EDX ECX SUB   0> IF   EBX INC
      BEGIN
         REPNZ SCASB   0= WHILE
         ECX PUSH   EDI PUSH
         EBX ESI MOV   EDX ECX MOV
         REPE CMPSB   0= NOT WHILE
         EDI POP   ECX POP
         REPEAT
         2 CELLS # ESP ADD   EAX EAX SUB
   THEN   THEN
   EDI 0 [EBP] MOV   EAX EBX MOV
   EDI POP   ESI POP
   RET   END-CODE

{ ------------------------------------------------------------------------
Case conversion

UPPER  converts a single character to uppercase.

UPCASE converts a string to uppercase. Note that it performs this
conversion in-situ, modifying the string given to it.
------------------------------------------------------------------------ }

ICODE UPPER ( char -- CHAR )
   CHAR a # BL CMP   U>= IF
      CHAR z # BL CMP   U<= IF
   32 # BL XOR   THEN THEN
   RET   END-CODE

CODE UPCASE ( a # -- )
   EBX ECX MOV   0 [EBP] EDX MOV
   ECXNZ IF   BEGIN
         0 [EDX] AL MOV   CHAR a # AL CMP   U>= IF
            CHAR z # AL CMP   U<= IF
               32 # AL XOR
         THEN THEN
         AL 0 [EDX] MOV   EDX INC
   LOOP THEN
   CELL [EBP] EBX MOV   2 CELLS [EBP] EBP LEA
   RET   END-CODE
