{ ====================================================================
Trap handler

Copyright 2020  FORTH, Inc.

Memory structures documented at the end of this file.
==================================================================== }

TARGET

{ ----------------------------------------------------------------------
Vectored exception handler

IOR_EXCEPT is the reserved throw code for Windows exceptions.
EXCEPTION# holds the ExceptionCode field received in ExceptionRecord.
TRACEBACK holds the register and stack contexts for the exception.

THROWBACK does the equivalent of IOR_WINEXCEPT THROW. The THROWBACK
code address is stuffed into the ContextRecord Eip field in
<EXCEPTION> below.

<EXCEPTION> is the SwiftForth PvectoredExceptionHandler (see
documentation below). It receives a pointer to ExceptionInfo on the
stack when called. It pulls the two pointers (ExceptionRecord and
ContextRecord) from ExceptionInfo, saving ExceptionCode in variable
EXCEPTION#, and copying the register and stack context to the
TRACEBACK buffer. EXCEPTION# and TRACEBACK are used by the throw code
debug display in the IDE. The address of THROWBACK replaces the EIP
value in ContextRecord and we return with EXCEPTION_CONTINUE_EXECUTION
so Windows will attempt to continue execution in that context, but
returning to THROWBACK instead, which does IOR_WINEXCEPT THROW back to
the CATCH of the code that caused the exception.

/EXCEPTION sets <EXCEPTION> as the vectored exception handler with the
flag 1 to prioritize this handler as first. This is called at startup.
---------------------------------------------------------------------- }

LABEL <EXCEPTION>
   4 [ESP] EDX MOV                      \ ExceptionInfo addr
   EBX PUSH   ESI PUSH   EDI PUSH       \ Save registers for return to Windows
   0 [EDX] EAX MOV   0 [EAX] EAX MOV    \ EAX = ExceptionCode
   4 [EDX] EBX MOV                      \ EBX = ContextRecord
   $F0000000 # EAX AND                  \ Mask off exception class
   $C0000000 # EAX CMP   0= NOT IF      \ Bail out if not Windows status code
      EDI POP   ESI POP   EBX POP
      EXCEPTION_CONTINUE_SEARCH # EAX MOV
   4 # RET   THEN
   0 [EDX] EAX MOV   0 [EAX] EAX MOV    \ EAX = ExceptionCode
   BEGIN ( *)  0 # EDX MOV              \ EDX = ORIGIN (filled in by /SIGNAL)
   EAX EXCEPTION# [EDX] MOV             \ Save exception code
   140 [EBX] ESI LEA                    \ ESI = start of registers in context record
   TRACEBACK [EDX] EDI LEA              \ EDI = traceback buffer in data space
   #TRACE_REGS # ECX MOV   REP MOVSD    \ Copy registers to traceback buffer
   196 [EBX] ESI MOV                    \ ESI = return stack addr ESP from context record
   #TRACE_RSTACK # ECX MOV   REP MOVSD  \ Copy return stack to traceback buffer
   180 [EBX] ESI MOV                    \ ESI = data stack addr EBP from context record
   #TRACE_DSTACK # ECX MOV   REP MOVSD  \ Copy return stack to traceback buffer
   'CATCHING [EDX] EAX LEA   ESP EDI MOV
   $8000 # ECX MOV   REPNE SCASD        \ Find CATCH stack frame
   0= IF   8 [EDI] EAX MOV              \ Get pushed ESI (user pointer)
   EAX 160 [EBX] MOV   THEN             \ Replace ESI in context record so THROW will work correctly
   THROWBACK [EDX] EAX LEA              \ Code address of THROWBACK
   EAX 184 [EBX] MOV                    \ Stuff THROWBACK addr in place of EIP in context record
   EDI POP   ESI POP   EBX POP          \ Restore registers
   EXCEPTION_CONTINUE_EXECUTION # EAX MOV
   4 # RET   END-CODE                   \ Return back to Windows and eventually to THROWBACK

( *) 1+ EQU 'ORIGIN                    \ Address of immediate value for ORIGIN

| : /SIGNAL ( -- )
   ORIGIN 'ORIGIN +ORIGIN !             \ Set our origin at ( *) above
   1 <EXCEPTION> +ORIGIN AddVectoredExceptionHandler DROP ;


\\ *********************************************************************

Function and structure documentation from Microsoft Win32 API:

<EXCEPTION> follows this function prototype:

LONG PvectoredExceptionHandler(
  _EXCEPTION_POINTERS *ExceptionInfo
)


ExceptionInfo structure passed to <EXCEPTION> handler by Windows:

typedef struct _EXCEPTION_POINTERS {
  PEXCEPTION_RECORD ExceptionRecord;
  PCONTEXT          ContextRecord;
} EXCEPTION_POINTERS, *PEXCEPTION_POINTERS;


Exception record structure pointed to by first cell of ExceptionRecord:

typedef struct _EXCEPTION_RECORD {
  DWORD                    ExceptionCode;
  DWORD                    ExceptionFlags;
  struct _EXCEPTION_RECORD *ExceptionRecord;
  PVOID                    ExceptionAddress;
  DWORD                    NumberParameters;
  ULONG_PTR                ExceptionInformation[EXCEPTION_MAXIMUM_PARAMETERS];
} EXCEPTION_RECORD;


Machine-dependent context record structure pointed to by second cell of ExceptionRecord:

typedef struct _CONTEXT {
    DWORD   ContextFlags;
    /* Debug registers */
    DWORD   Dr0;
    DWORD   Dr1;
    DWORD   Dr2;
    DWORD   Dr3;
    DWORD   Dr6;
    DWORD   Dr7;
    /* Floating-point context */
    DWORD   ControlWord;
    DWORD   StatusWord;
    DWORD   TagWord;
    DWORD   ErrorOffset;
    DWORD   ErrorSelector;
    DWORD   DataOffset;
    DWORD   DataSelector;
    BYTE    RegisterArea[80];
    DWORD   Cr0NpxState;
    /* CPU registers -- offset=140 */
    DWORD   SegGs;      // 140
    DWORD   SegFs;      // 144
    DWORD   SegEs;      // 148
    DWORD   SegDs;      // 152
    DWORD   Edi;        // 156
    DWORD   Esi;        // 160
    DWORD   Ebx;        // 164
    DWORD   Edx;        // 168
    DWORD   Ecx;        // 172
    DWORD   Eax;        // 176
    DWORD   Ebp;        // 180
    DWORD   Eip;        // 184
    DWORD   SegCs;      // 188
    DWORD   EFlags;     // 192
    DWORD   Esp;        // 196
    DWORD   SegSs;      // 200
    BYTE    ExtendedRegisters[512];
} CONTEXT;
